#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef _MPI
	#include <mpi.h>
#endif
#include "lis.h"
#include "lis_matvec.h"


LIS_MATVEC_DIA lis_matvec_dia_xxx[5][8] = { 
	{
		lis_matvec_dia   , lis_matvec_dia_u2,lis_matvec_dia_u3 ,lis_matvec_dia_u4,
		lis_matvec_dia_u5, lis_matvec_dia_u6,lis_matvec_dia_u7 ,lis_matvec_dia_u8,
	},
	{
		lis_matvec_dia   ,lis_matvec_dia_p2 ,lis_matvec_dia_p3 ,lis_matvec_dia_p4,
		lis_matvec_dia_p5,lis_matvec_dia_p6 ,lis_matvec_dia_p7 ,lis_matvec_dia_p8,
	},
	{
		lis_matvec_dia_u1s, lis_matvec_dia_u2s,lis_matvec_dia_u3s ,lis_matvec_dia_u4s,
		lis_matvec_dia_u5s, lis_matvec_dia_u6s,lis_matvec_dia_u7s ,lis_matvec_dia_u8s,
	},
	{
		lis_matvec_dia_p1s,lis_matvec_dia_p2s ,lis_matvec_dia_p3s ,lis_matvec_dia_p4s,
		lis_matvec_dia_p5s,lis_matvec_dia_p6s ,lis_matvec_dia_p7s ,lis_matvec_dia_p8s,
	},
	{
		/*
		lis_matvec_dia_u4s_sse2,lis_matvec_dia_u4s_sse2 ,lis_matvec_dia_u4s_sse2 ,lis_matvec_dia_u4s_sse2,
		lis_matvec_dia_u6s_sse2,lis_matvec_dia_u6s_sse2 ,lis_matvec_dia_u8s_sse2 ,lis_matvec_dia_u8s_sse2,
		*/
		NULL,NULL ,NULL ,NULL,
		NULL,NULL ,NULL ,NULL,
	}
};

/************************/
/* Unrolling            */
/************************/
void lis_matvec_dia_u2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2;
	LIS_SCALAR *x1,*x2;
	LIS_SCALAR *v1,*v2;

	n    = A->n;
	nnd  = A->nnd;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,n-jj);
		kk = j*n;

		yy  = &y[js];
		xx  = &x[jj+js];
		vv  = &A->value[kk+js];
		for(i=js+1;i<je;i+=2)
		{
			y1 = yy++; x1 = xx++; v1 = vv++;
			y2 = yy++; x2 = xx++; v2 = vv++;
			*y1 += *v1 * *x1;
			*y2 += *v2 * *x2;
		}
		for(i-=1;i<je;i++)
		{ 
			*yy += *vv * *xx;
			yy++; xx++; vv++;
		}
	}
}

void lis_matvec_dia_u3(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3;
	LIS_SCALAR *x1,*x2,*x3;
	LIS_SCALAR *v1,*v2,*v3;

	n    = A->n;
	nnd  = A->nnd;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,n-jj);
		kk = j*n;

		yy  = &y[js];
		xx  = &x[jj+js];
		vv  = &A->value[kk+js];
		for(i=js+2;i<je;i+=3)
		{
			y1 = yy++; x1 = xx++; v1 = vv++;
			y2 = yy++; x2 = xx++; v2 = vv++;
			y3 = yy++; x3 = xx++; v3 = vv++;
			*y1 += *v1 * *x1;
			*y2 += *v2 * *x2;
			*y3 += *v3 * *x3;
		}
		for(i-=2;i<je;i++)
		{ 
			*yy += *vv * *xx;
			yy++; xx++; vv++;
		}
	}
}

void lis_matvec_dia_u4(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4;
	LIS_SCALAR *x1,*x2,*x3,*x4;
	LIS_SCALAR *v1,*v2,*v3,*v4;

	n    = A->n;
	nnd  = A->nnd;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,n-jj);
		kk = j*n;

		yy  = &y[js];
		xx  = &x[jj+js];
		vv  = &A->value[kk+js];
		for(i=js+3;i<je;i+=4)
		{
			y1 = yy++; x1 = xx++; v1 = vv++;
			y2 = yy++; x2 = xx++; v2 = vv++;
			y3 = yy++; x3 = xx++; v3 = vv++;
			y4 = yy++; x4 = xx++; v4 = vv++;
			*y1 += *v1 * *x1;
			*y2 += *v2 * *x2;
			*y3 += *v3 * *x3;
			*y4 += *v4 * *x4;
		}
		for(i-=3;i<je;i++)
		{ 
			*yy += *vv * *xx;
			yy++; xx++; vv++;
		}
	}
}

void lis_matvec_dia_u5(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4,*y5;
	LIS_SCALAR *x1,*x2,*x3,*x4,*x5;
	LIS_SCALAR *v1,*v2,*v3,*v4,*v5;

	n    = A->n;
	nnd  = A->nnd;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,n-jj);
		kk = j*n;

		yy  = &y[js];
		xx  = &x[jj+js];
		vv  = &A->value[kk+js];
		for(i=js+4;i<je;i+=5)
		{
			y1 = yy++; x1 = xx++; v1 = vv++;
			y2 = yy++; x2 = xx++; v2 = vv++;
			y3 = yy++; x3 = xx++; v3 = vv++;
			y4 = yy++; x4 = xx++; v4 = vv++;
			y5 = yy++; x5 = xx++; v5 = vv++;
			*y1 += *v1 * *x1;
			*y2 += *v2 * *x2;
			*y3 += *v3 * *x3;
			*y4 += *v4 * *x4;
			*y5 += *v5 * *x5;
		}
		for(i-=4;i<je;i++)
		{ 
			*yy += *vv * *xx;
			yy++; xx++; vv++;
		}
	}
}

void lis_matvec_dia_u6(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4,*y5,*y6;
	LIS_SCALAR *x1,*x2,*x3,*x4,*x5,*x6;
	LIS_SCALAR *v1,*v2,*v3,*v4,*v5,*v6;

	n    = A->n;
	nnd  = A->nnd;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,n-jj);
		kk = j*n;

		yy  = &y[js];
		xx  = &x[jj+js];
		vv  = &A->value[kk+js];
		for(i=js+5;i<je;i+=6)
		{
			y1 = yy++; x1 = xx++; v1 = vv++;
			y2 = yy++; x2 = xx++; v2 = vv++;
			y3 = yy++; x3 = xx++; v3 = vv++;
			y4 = yy++; x4 = xx++; v4 = vv++;
			y5 = yy++; x5 = xx++; v5 = vv++;
			y6 = yy++; x6 = xx++; v6 = vv++;
			*y1 += *v1 * *x1;
			*y2 += *v2 * *x2;
			*y3 += *v3 * *x3;
			*y4 += *v4 * *x4;
			*y5 += *v5 * *x5;
			*y6 += *v6 * *x6;
		}
		for(i-=5;i<je;i++)
		{ 
			*yy += *vv * *xx;
			yy++; xx++; vv++;
		}
	}
}

void lis_matvec_dia_u7(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4,*y5,*y6,*y7;
	LIS_SCALAR *x1,*x2,*x3,*x4,*x5,*x6,*x7;
	LIS_SCALAR *v1,*v2,*v3,*v4,*v5,*v6,*v7;

	n    = A->n;
	nnd  = A->nnd;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,n-jj);
		kk = j*n;

		yy  = &y[js];
		xx  = &x[jj+js];
		vv  = &A->value[kk+js];
		for(i=js+6;i<je;i+=7)
		{
			y1 = yy++; x1 = xx++; v1 = vv++;
			y2 = yy++; x2 = xx++; v2 = vv++;
			y3 = yy++; x3 = xx++; v3 = vv++;
			y4 = yy++; x4 = xx++; v4 = vv++;
			y5 = yy++; x5 = xx++; v5 = vv++;
			y6 = yy++; x6 = xx++; v6 = vv++;
			y7 = yy++; x7 = xx++; v7 = vv++;
			*y1 += *v1 * *x1;
			*y2 += *v2 * *x2;
			*y3 += *v3 * *x3;
			*y4 += *v4 * *x4;
			*y5 += *v5 * *x5;
			*y6 += *v6 * *x6;
			*y7 += *v7 * *x7;
		}
		for(i-=6;i<je;i++)
		{ 
			*yy += *vv * *xx;
			yy++; xx++; vv++;
		}
	}
}

void lis_matvec_dia_u8(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4,*y5,*y6,*y7,*y8;
	LIS_SCALAR *x1,*x2,*x3,*x4,*x5,*x6,*x7,*x8;
	LIS_SCALAR *v1,*v2,*v3,*v4,*v5,*v6,*v7,*v8;

	n    = A->n;
	nnd  = A->nnd;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,n-jj);
		kk = j*n;

		yy  = &y[js];
		xx  = &x[jj+js];
		vv  = &A->value[kk+js];
		for(i=js+7;i<je;i+=8)
		{
			y1 = yy++; x1 = xx++; v1 = vv++;
			y2 = yy++; x2 = xx++; v2 = vv++;
			y3 = yy++; x3 = xx++; v3 = vv++;
			y4 = yy++; x4 = xx++; v4 = vv++;
			y5 = yy++; x5 = xx++; v5 = vv++;
			y6 = yy++; x6 = xx++; v6 = vv++;
			y7 = yy++; x7 = xx++; v7 = vv++;
			y8 = yy++; x8 = xx++; v8 = vv++;
			*y1 += *v1 * *x1;
			*y2 += *v2 * *x2;
			*y3 += *v3 * *x3;
			*y4 += *v4 * *x4;
			*y5 += *v5 * *x5;
			*y6 += *v6 * *x6;
			*y7 += *v7 * *x7;
			*y8 += *v8 * *x8;
		}
		for(i-=7;i<je;i++)
		{ 
			*yy += *vv * *xx;
			yy++; xx++; vv++;
		}
	}
}

/************************/
/* Pipeline             */
/************************/
void lis_matvec_dia_p2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2;
	LIS_SCALAR *x1,*x2;
	LIS_SCALAR *v1,*v2;

	n    = A->n;
	nnd  = A->nnd;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,n-jj);
		kk = j*n;
		yy = &y[js];
		xx = &x[jj+js];
		vv = &A->value[kk+js];

		if( je-js<3 )
		{
			i = js;
		}
		else
		{
			y1 = yy++; x1 = xx++; v1 = vv++;
			for(i=js+2;i<je;i+=2)
			{
				y2 = yy++; x2 = xx++; v2 = vv++;
				*y1 += *v1 * *x1;

				y1 = yy++; x1 = xx++; v1 = vv++;
				*y2 += *v2 * *x2;
			}
			*y1 += *v1 * *x1;
			i-=1;
		}
		for(;i<je;i++)
		{ 
			*yy += *vv * *xx;
			yy++; xx++; vv++;
		}
	}
}

void lis_matvec_dia_p3(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3;
	LIS_SCALAR *x1,*x2,*x3;
	LIS_SCALAR *v1,*v2,*v3;

	n    = A->n;
	nnd  = A->nnd;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,n-jj);
		kk = j*n;
		yy = &y[js];
		xx = &x[jj+js];
		vv = &A->value[kk+js];

		if( je-js<4 )
		{
			i = js;
		}
		else
		{
			y1 = yy++; x1 = xx++; v1 = vv++;
			y2 = yy++; x2 = xx++; v2 = vv++;
			for(i=js+4;i<je;i+=3)
			{
				y3 = yy++; x3 = xx++; v3 = vv++;
				*y1 += *v1 * *x1;

				y1 = yy++; x1 = xx++; v1 = vv++;
				*y2 += *v2 * *x2;

				y2 = yy++; x2 = xx++; v2 = vv++;
				*y3 += *v3 * *x3;
			}
			*y1 += *v1 * *x1;
			*y2 += *v2 * *x2;
			i-=2;
		}
		for(;i<je;i++)
		{ 
			*yy += *vv * *xx;
			yy++; xx++; vv++;
		}
	}
}

void lis_matvec_dia_p4(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4;
	LIS_SCALAR *x1,*x2,*x3,*x4;
	LIS_SCALAR *v1,*v2,*v3,*v4;

	n    = A->n;
	nnd  = A->nnd;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,n-jj);
		kk = j*n;
		yy = &y[js];
		xx = &x[jj+js];
		vv = &A->value[kk+js];

		if( je-js<5 )
		{
			i = js;
		}
		else
		{
			y1 = yy++; x1 = xx++; v1 = vv++;
			y2 = yy++; x2 = xx++; v2 = vv++;
			y3 = yy++; x3 = xx++; v3 = vv++;
			for(i=js+6;i<je;i+=4)
			{
				y4 = yy++; x4 = xx++; v4 = vv++;
				*y1 += *v1 * *x1;

				y1 = yy++; x1 = xx++; v1 = vv++;
				*y2 += *v2 * *x2;

				y2 = yy++; x2 = xx++; v2 = vv++;
				*y3 += *v3 * *x3;

				y3 = yy++; x3 = xx++; v3 = vv++;
				*y4 += *v4 * *x4;
			}
			*y1 += *v1 * *x1;
			*y2 += *v2 * *x2;
			*y3 += *v3 * *x3;
			i-=3;
		}
		for(;i<je;i++)
		{ 
			*yy += *vv * *xx;
			yy++; xx++; vv++;
		}
	}
}

void lis_matvec_dia_p5(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4,*y5;
	LIS_SCALAR *x1,*x2,*x3,*x4,*x5;
	LIS_SCALAR *v1,*v2,*v3,*v4,*v5;

	n    = A->n;
	nnd  = A->nnd;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,n-jj);
		kk = j*n;
		yy = &y[js];
		xx = &x[jj+js];
		vv = &A->value[kk+js];

		if( je-js<6 )
		{
			i = js;
		}
		else
		{
			y1 = yy++; x1 = xx++; v1 = vv++;
			y2 = yy++; x2 = xx++; v2 = vv++;
			y3 = yy++; x3 = xx++; v3 = vv++;
			y4 = yy++; x4 = xx++; v4 = vv++;
			for(i=js+8;i<je;i+=5)
			{
				y5 = yy++; x5 = xx++; v5 = vv++;
				*y1 += *v1 * *x1;

				y1 = yy++; x1 = xx++; v1 = vv++;
				*y2 += *v2 * *x2;

				y2 = yy++; x2 = xx++; v2 = vv++;
				*y3 += *v3 * *x3;

				y3 = yy++; x3 = xx++; v3 = vv++;
				*y4 += *v4 * *x4;

				y4 = yy++; x4 = xx++; v4 = vv++;
				*y5 += *v5 * *x5;
			}
			*y1 += *v1 * *x1;
			*y2 += *v2 * *x2;
			*y3 += *v3 * *x3;
			*y4 += *v4 * *x4;
			i-=4;
		}
		for(;i<je;i++)
		{ 
			*yy += *vv * *xx;
			yy++; xx++; vv++;
		}
	}
}

void lis_matvec_dia_p6(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4,*y5,*y6;
	LIS_SCALAR *x1,*x2,*x3,*x4,*x5,*x6;
	LIS_SCALAR *v1,*v2,*v3,*v4,*v5,*v6;

	n    = A->n;
	nnd  = A->nnd;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,n-jj);
		kk = j*n;
		yy = &y[js];
		xx = &x[jj+js];
		vv = &A->value[kk+js];

		if( je-js<7 )
		{
			i = js;
		}
		else
		{
			y1 = yy++; x1 = xx++; v1 = vv++;
			y2 = yy++; x2 = xx++; v2 = vv++;
			y3 = yy++; x3 = xx++; v3 = vv++;
			y4 = yy++; x4 = xx++; v4 = vv++;
			y5 = yy++; x5 = xx++; v5 = vv++;
			for(i=js+10;i<je;i+=6)
			{
				y6 = yy++; x6 = xx++; v6 = vv++;
				*y1 += *v1 * *x1;

				y1 = yy++; x1 = xx++; v1 = vv++;
				*y2 += *v2 * *x2;

				y2 = yy++; x2 = xx++; v2 = vv++;
				*y3 += *v3 * *x3;

				y3 = yy++; x3 = xx++; v3 = vv++;
				*y4 += *v4 * *x4;

				y4 = yy++; x4 = xx++; v4 = vv++;
				*y5 += *v5 * *x5;

				y5 = yy++; x5 = xx++; v5 = vv++;
				*y6 += *v6 * *x6;
			}
			*y1 += *v1 * *x1;
			*y2 += *v2 * *x2;
			*y3 += *v3 * *x3;
			*y4 += *v4 * *x4;
			*y5 += *v5 * *x5;
			i-=5;
		}
		for(;i<je;i++)
		{ 
			*yy += *vv * *xx;
			yy++; xx++; vv++;
		}
	}
}

void lis_matvec_dia_p7(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4,*y5,*y6,*y7;
	LIS_SCALAR *x1,*x2,*x3,*x4,*x5,*x6,*x7;
	LIS_SCALAR *v1,*v2,*v3,*v4,*v5,*v6,*v7;

	n    = A->n;
	nnd  = A->nnd;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,n-jj);
		kk = j*n;
		yy = &y[js];
		xx = &x[jj+js];
		vv = &A->value[kk+js];

		if( je-js<8 )
		{
			i = js;
		}
		else
		{
			y1 = yy++; x1 = xx++; v1 = vv++;
			y2 = yy++; x2 = xx++; v2 = vv++;
			y3 = yy++; x3 = xx++; v3 = vv++;
			y4 = yy++; x4 = xx++; v4 = vv++;
			y5 = yy++; x5 = xx++; v5 = vv++;
			y6 = yy++; x6 = xx++; v6 = vv++;
			for(i=js+12;i<je;i+=7)
			{
				y7 = yy++; x7 = xx++; v7 = vv++;
				*y1 += *v1 * *x1;

				y1 = yy++; x1 = xx++; v1 = vv++;
				*y2 += *v2 * *x2;

				y2 = yy++; x2 = xx++; v2 = vv++;
				*y3 += *v3 * *x3;

				y3 = yy++; x3 = xx++; v3 = vv++;
				*y4 += *v4 * *x4;

				y4 = yy++; x4 = xx++; v4 = vv++;
				*y5 += *v5 * *x5;

				y5 = yy++; x5 = xx++; v5 = vv++;
				*y6 += *v6 * *x6;

				y6 = yy++; x6 = xx++; v6 = vv++;
				*y7 += *v7 * *x7;
			}
			*y1 += *v1 * *x1;
			*y2 += *v2 * *x2;
			*y3 += *v3 * *x3;
			*y4 += *v4 * *x4;
			*y5 += *v5 * *x5;
			*y6 += *v6 * *x6;
			i-=6;
		}
		for(;i<je;i++)
		{ 
			*yy += *vv * *xx;
			yy++; xx++; vv++;
		}
	}
}

void lis_matvec_dia_p8(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4,*y5,*y6,*y7,*y8;
	LIS_SCALAR *x1,*x2,*x3,*x4,*x5,*x6,*x7,*x8;
	LIS_SCALAR *v1,*v2,*v3,*v4,*v5,*v6,*v7,*v8;

	n    = A->n;
	nnd  = A->nnd;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(j=0;j<nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,n-jj);
		kk = j*n;
		yy = &y[js];
		xx = &x[jj+js];
		vv = &A->value[kk+js];

		if( je-js<9 )
		{
			i = js;
		}
		else
		{
			y1 = yy++; x1 = xx++; v1 = vv++;
			y2 = yy++; x2 = xx++; v2 = vv++;
			y3 = yy++; x3 = xx++; v3 = vv++;
			y4 = yy++; x4 = xx++; v4 = vv++;
			y5 = yy++; x5 = xx++; v5 = vv++;
			y6 = yy++; x6 = xx++; v6 = vv++;
			y7 = yy++; x7 = xx++; v7 = vv++;
			for(i=js+14;i<je;i+=8)
			{
				y8 = yy++; x8 = xx++; v8 = vv++;
				*y1 += *v1 * *x1;

				y1 = yy++; x1 = xx++; v1 = vv++;
				*y2 += *v2 * *x2;

				y2 = yy++; x2 = xx++; v2 = vv++;
				*y3 += *v3 * *x3;

				y3 = yy++; x3 = xx++; v3 = vv++;
				*y4 += *v4 * *x4;

				y4 = yy++; x4 = xx++; v4 = vv++;
				*y5 += *v5 * *x5;

				y5 = yy++; x5 = xx++; v5 = vv++;
				*y6 += *v6 * *x6;

				y6 = yy++; x6 = xx++; v6 = vv++;
				*y7 += *v7 * *x7;

				y7 = yy++; x7 = xx++; v7 = vv++;
				*y8 += *v8 * *x8;
			}
			*y1 += *v1 * *x1;
			*y2 += *v2 * *x2;
			*y3 += *v3 * *x3;
			*y4 += *v4 * *x4;
			*y5 += *v5 * *x5;
			*y6 += *v6 * *x6;
			*y7 += *v7 * *x7;
			i-=7;
		}
		for(;i<je;i++)
		{ 
			*yy += *vv * *xx;
			yy++; xx++; vv++;
		}
	}
}

/*****************************/
/* Unrolling + Strip-mining  */
/*****************************/
void lis_matvec_dia_u1s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}

	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			kk = j*n;
			yy = &y[js];
			xx = &x[jj+js];
			vv = &A->value[kk+js];
			for(i=js;i<je;i++)
			{
				*yy += *vv * *xx;
				yy++; xx++; vv++;
			}
		}
	}
}

void lis_matvec_dia_u2s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2;
	LIS_SCALAR *x1,*x2;
	LIS_SCALAR *v1,*v2;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}

	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			kk = j*n;
			yy = &y[js];
			xx = &x[jj+js];
			vv = &A->value[kk+js];
			for(i=js+1;i<je;i+=2)
			{
				y1 = yy++; x1 = xx++; v1 = vv++;
				y2 = yy++; x2 = xx++; v2 = vv++;
				*y1 += *v1 * *x1;
				*y2 += *v2 * *x2;
			}
			for(i-=1;i<je;i++)
			{ 
				*yy += *vv * *xx;
				yy++; xx++; vv++;
			}
		}
	}
}

void lis_matvec_dia_u3s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3;
	LIS_SCALAR *x1,*x2,*x3;
	LIS_SCALAR *v1,*v2,*v3;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			kk = j*n;
			yy = &y[js];
			xx = &x[jj+js];
			vv = &A->value[kk+js];
			for(i=js+2;i<je;i+=3)
			{
				y1 = yy++; x1 = xx++; v1 = vv++;
				y2 = yy++; x2 = xx++; v2 = vv++;
				y3 = yy++; x3 = xx++; v3 = vv++;
				*y1 += *v1 * *x1;
				*y2 += *v2 * *x2;
				*y3 += *v3 * *x3;
			}
			for(i-=2;i<je;i++)
			{ 
				*yy += *vv * *xx;
				yy++; xx++; vv++;
			}
		}
	}
}

void lis_matvec_dia_u4s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4;
	LIS_SCALAR *x1,*x2,*x3,*x4;
	LIS_SCALAR *v1,*v2,*v3,*v4;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			kk = j*n;
			yy = &y[js];
			xx = &x[jj+js];
			vv = &A->value[kk+js];
			for(i=js+3;i<je;i+=4)
			{
				y1 = yy++; x1 = xx++; v1 = vv++;
				y2 = yy++; x2 = xx++; v2 = vv++;
				y3 = yy++; x3 = xx++; v3 = vv++;
				y4 = yy++; x4 = xx++; v4 = vv++;
				*y1 += *v1 * *x1;
				*y2 += *v2 * *x2;
				*y3 += *v3 * *x3;
				*y4 += *v4 * *x4;
			}
			for(i-=3;i<je;i++)
			{ 
				*yy += *vv * *xx;
				yy++; xx++; vv++;
			}
		}
	}
}

void lis_matvec_dia_u5s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4,*y5;
	LIS_SCALAR *x1,*x2,*x3,*x4,*x5;
	LIS_SCALAR *v1,*v2,*v3,*v4,*v5;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			kk = j*n;
			yy = &y[js];
			xx = &x[jj+js];
			vv = &A->value[kk+js];
			for(i=js+4;i<je;i+=5)
			{
				y1 = yy++; x1 = xx++; v1 = vv++;
				y2 = yy++; x2 = xx++; v2 = vv++;
				y3 = yy++; x3 = xx++; v3 = vv++;
				y4 = yy++; x4 = xx++; v4 = vv++;
				y5 = yy++; x5 = xx++; v5 = vv++;
				*y1 += *v1 * *x1;
				*y2 += *v2 * *x2;
				*y3 += *v3 * *x3;
				*y4 += *v4 * *x4;
				*y5 += *v5 * *x5;
			}
			for(i-=4;i<je;i++)
			{ 
				*yy += *vv * *xx;
				yy++; xx++; vv++;
			}
		}
	}
}

void lis_matvec_dia_u6s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4,*y5,*y6;
	LIS_SCALAR *x1,*x2,*x3,*x4,*x5,*x6;
	LIS_SCALAR *v1,*v2,*v3,*v4,*v5,*v6;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			kk = j*n;
			yy = &y[js];
			xx = &x[jj+js];
			vv = &A->value[kk+js];
			for(i=js+5;i<je;i+=6)
			{
				y1 = yy++; x1 = xx++; v1 = vv++;
				y2 = yy++; x2 = xx++; v2 = vv++;
				y3 = yy++; x3 = xx++; v3 = vv++;
				y4 = yy++; x4 = xx++; v4 = vv++;
				y5 = yy++; x5 = xx++; v5 = vv++;
				y6 = yy++; x6 = xx++; v6 = vv++;
				*y1 += *v1 * *x1;
				*y2 += *v2 * *x2;
				*y3 += *v3 * *x3;
				*y4 += *v4 * *x4;
				*y5 += *v5 * *x5;
				*y6 += *v6 * *x6;
			}
			for(i-=5;i<je;i++)
			{ 
				*yy += *vv * *xx;
				yy++; xx++; vv++;
			}
		}
	}
}

void lis_matvec_dia_u7s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4,*y5,*y6,*y7;
	LIS_SCALAR *x1,*x2,*x3,*x4,*x5,*x6,*x7;
	LIS_SCALAR *v1,*v2,*v3,*v4,*v5,*v6,*v7;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			kk = j*n;
			yy = &y[js];
			xx = &x[jj+js];
			vv = &A->value[kk+js];
			for(i=js+6;i<je;i+=7)
			{
				y1 = yy++; x1 = xx++; v1 = vv++;
				y2 = yy++; x2 = xx++; v2 = vv++;
				y3 = yy++; x3 = xx++; v3 = vv++;
				y4 = yy++; x4 = xx++; v4 = vv++;
				y5 = yy++; x5 = xx++; v5 = vv++;
				y6 = yy++; x6 = xx++; v6 = vv++;
				y7 = yy++; x7 = xx++; v7 = vv++;
				*y1 += *v1 * *x1;
				*y2 += *v2 * *x2;
				*y3 += *v3 * *x3;
				*y4 += *v4 * *x4;
				*y5 += *v5 * *x5;
				*y6 += *v6 * *x6;
				*y7 += *v7 * *x7;
			}
			for(i-=6;i<je;i++)
			{ 
				*yy += *vv * *xx;
				yy++; xx++; vv++;
			}
		}
	}
}

void lis_matvec_dia_u8s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4,*y5,*y6,*y7,*y8;
	LIS_SCALAR *x1,*x2,*x3,*x4,*x5,*x6,*x7,*x8;
	LIS_SCALAR *v1,*v2,*v3,*v4,*v5,*v6,*v7,*v8;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			kk = j*n;
			yy = &y[js];
			xx = &x[jj+js];
			vv = &A->value[kk+js];
			for(i=js+7;i<je;i+=8)
			{
				y1 = yy++; x1 = xx++; v1 = vv++;
				y2 = yy++; x2 = xx++; v2 = vv++;
				y3 = yy++; x3 = xx++; v3 = vv++;
				y4 = yy++; x4 = xx++; v4 = vv++;
				y5 = yy++; x5 = xx++; v5 = vv++;
				y6 = yy++; x6 = xx++; v6 = vv++;
				y7 = yy++; x7 = xx++; v7 = vv++;
				y8 = yy++; x8 = xx++; v8 = vv++;
				*y1 += *v1 * *x1;
				*y2 += *v2 * *x2;
				*y3 += *v3 * *x3;
				*y4 += *v4 * *x4;
				*y5 += *v5 * *x5;
				*y6 += *v6 * *x6;
				*y7 += *v7 * *x7;
				*y8 += *v8 * *x8;
			}
			for(i-=7;i<je;i++)
			{ 
				*yy += *vv * *xx;
				yy++; xx++; vv++;
			}
		}
	}
}

void lis_matvec_dia_p1s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1;
	LIS_SCALAR *x1;
	LIS_SCALAR *v1;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			kk = j*n;
			yy = &y[js];
			xx = &x[jj+js];
			vv = &A->value[kk+js];

			y1 = yy++; x1 = xx++; v1 = vv++;
			for(i=js+1;i<je;i++)
			{
				*y1 += *v1 * *x1;
				y1 = yy++; x1 = xx++; v1 = vv++;
			}
			*y1 += *v1 * *x1;
		}
	}
}

void lis_matvec_dia_p2s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2;
	LIS_SCALAR *x1,*x2;
	LIS_SCALAR *v1,*v2;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			kk = j*n;
			yy = &y[js];
			xx = &x[jj+js];
			vv = &A->value[kk+js];

			if( je-js<3 )
			{
				i = js;
			}
			else
			{
				y1 = yy++; x1 = xx++; v1 = vv++;
				for(i=js+2;i<je;i+=2)
				{
					y2 = yy++; x2 = xx++; v2 = vv++;
					*y1 += *v1 * *x1;

					y1 = yy++; x1 = xx++; v1 = vv++;
					*y2 += *v2 * *x2;
				}
				*y1 += *v1 * *x1;
				i-=1;
			}
			for(;i<je;i++)
			{ 
				*yy += *vv * *xx;
				yy++; xx++; vv++;
			}
		}
	}
}

void lis_matvec_dia_p3s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3;
	LIS_SCALAR *x1,*x2,*x3;
	LIS_SCALAR *v1,*v2,*v3;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			kk = j*n;
			yy = &y[js];
			xx = &x[jj+js];
			vv = &A->value[kk+js];

			if( je-js<4 )
			{
				i = js;
			}
			else
			{
				y1 = yy++; x1 = xx++; v1 = vv++;
				y2 = yy++; x2 = xx++; v2 = vv++;
				for(i=js+4;i<je;i+=3)
				{
					y3 = yy++; x3 = xx++; v3 = vv++;
					*y1 += *v1 * *x1;

					y1 = yy++; x1 = xx++; v1 = vv++;
					*y2 += *v2 * *x2;

					y2 = yy++; x2 = xx++; v2 = vv++;
					*y3 += *v3 * *x3;
				}
				*y1 += *v1 * *x1;
				*y2 += *v2 * *x2;
				i-=2;
			}
			for(;i<je;i++)
			{ 
				*yy += *vv * *xx;
				yy++; xx++; vv++;
			}
		}
	}
}

void lis_matvec_dia_p4s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4;
	LIS_SCALAR *x1,*x2,*x3,*x4;
	LIS_SCALAR *v1,*v2,*v3,*v4;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			kk = j*n;
			yy = &y[js];
			xx = &x[jj+js];
			vv = &A->value[kk+js];

			if( je-js<5 )
			{
				i = js;
			}
			else
			{
				y1 = yy++; x1 = xx++; v1 = vv++;
				y2 = yy++; x2 = xx++; v2 = vv++;
				y3 = yy++; x3 = xx++; v3 = vv++;
				for(i=js+6;i<je;i+=4)
				{
					y4 = yy++; x4 = xx++; v4 = vv++;
					*y1 += *v1 * *x1;

					y1 = yy++; x1 = xx++; v1 = vv++;
					*y2 += *v2 * *x2;

					y2 = yy++; x2 = xx++; v2 = vv++;
					*y3 += *v3 * *x3;

					y3 = yy++; x3 = xx++; v3 = vv++;
					*y4 += *v4 * *x4;
				}
				*y1 += *v1 * *x1;
				*y2 += *v2 * *x2;
				*y3 += *v3 * *x3;
				i-=3;
			}
			for(;i<je;i++)
			{ 
				*yy += *vv * *xx;
				yy++; xx++; vv++;
			}
		}
	}
}

void lis_matvec_dia_p5s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4,*y5;
	LIS_SCALAR *x1,*x2,*x3,*x4,*x5;
	LIS_SCALAR *v1,*v2,*v3,*v4,*v5;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			kk = j*n;
			yy = &y[js];
			xx = &x[jj+js];
			vv = &A->value[kk+js];

			if( je-js<6 )
			{
				i = js;
			}
			else
			{
				y1 = yy++; x1 = xx++; v1 = vv++;
				y2 = yy++; x2 = xx++; v2 = vv++;
				y3 = yy++; x3 = xx++; v3 = vv++;
				y4 = yy++; x4 = xx++; v4 = vv++;
				for(i=js+8;i<je;i+=5)
				{
					y5 = yy++; x5 = xx++; v5 = vv++;
					*y1 += *v1 * *x1;

					y1 = yy++; x1 = xx++; v1 = vv++;
					*y2 += *v2 * *x2;

					y2 = yy++; x2 = xx++; v2 = vv++;
					*y3 += *v3 * *x3;

					y3 = yy++; x3 = xx++; v3 = vv++;
					*y4 += *v4 * *x4;

					y4 = yy++; x4 = xx++; v4 = vv++;
					*y5 += *v5 * *x5;
				}
				*y1 += *v1 * *x1;
				*y2 += *v2 * *x2;
				*y3 += *v3 * *x3;
				*y4 += *v4 * *x4;
				i-=4;
			}
			for(;i<je;i++)
			{ 
				*yy += *vv * *xx;
				yy++; xx++; vv++;
			}
		}
	}
}

void lis_matvec_dia_p6s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4,*y5,*y6;
	LIS_SCALAR *x1,*x2,*x3,*x4,*x5,*x6;
	LIS_SCALAR *v1,*v2,*v3,*v4,*v5,*v6;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			kk = j*n;
			yy = &y[js];
			xx = &x[jj+js];
			vv = &A->value[kk+js];

			if( je-js<7 )
			{
				i = js;
			}
			else
			{
				y1 = yy++; x1 = xx++; v1 = vv++;
				y2 = yy++; x2 = xx++; v2 = vv++;
				y3 = yy++; x3 = xx++; v3 = vv++;
				y4 = yy++; x4 = xx++; v4 = vv++;
				y5 = yy++; x5 = xx++; v5 = vv++;
				for(i=js+10;i<je;i+=6)
				{
					y6 = yy++; x6 = xx++; v6 = vv++;
					*y1 += *v1 * *x1;

					y1 = yy++; x1 = xx++; v1 = vv++;
					*y2 += *v2 * *x2;

					y2 = yy++; x2 = xx++; v2 = vv++;
					*y3 += *v3 * *x3;

					y3 = yy++; x3 = xx++; v3 = vv++;
					*y4 += *v4 * *x4;

					y4 = yy++; x4 = xx++; v4 = vv++;
					*y5 += *v5 * *x5;

					y5 = yy++; x5 = xx++; v5 = vv++;
					*y6 += *v6 * *x6;
				}
				*y1 += *v1 * *x1;
				*y2 += *v2 * *x2;
				*y3 += *v3 * *x3;
				*y4 += *v4 * *x4;
				*y5 += *v5 * *x5;
				i-=5;
			}
			for(;i<je;i++)
			{ 
				*yy += *vv * *xx;
				yy++; xx++; vv++;
			}
		}
	}
}

void lis_matvec_dia_p7s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4,*y5,*y6,*y7;
	LIS_SCALAR *x1,*x2,*x3,*x4,*x5,*x6,*x7;
	LIS_SCALAR *v1,*v2,*v3,*v4,*v5,*v6,*v7;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			kk = j*n;
			yy = &y[js];
			xx = &x[jj+js];
			vv = &A->value[kk+js];

			if( je-js<8 )
			{
				i = js;
			}
			else
			{
				y1 = yy++; x1 = xx++; v1 = vv++;
				y2 = yy++; x2 = xx++; v2 = vv++;
				y3 = yy++; x3 = xx++; v3 = vv++;
				y4 = yy++; x4 = xx++; v4 = vv++;
				y5 = yy++; x5 = xx++; v5 = vv++;
				y6 = yy++; x6 = xx++; v6 = vv++;
				for(i=js+12;i<je;i+=7)
				{
					y7 = yy++; x7 = xx++; v7 = vv++;
					*y1 += *v1 * *x1;

					y1 = yy++; x1 = xx++; v1 = vv++;
					*y2 += *v2 * *x2;

					y2 = yy++; x2 = xx++; v2 = vv++;
					*y3 += *v3 * *x3;

					y3 = yy++; x3 = xx++; v3 = vv++;
					*y4 += *v4 * *x4;

					y4 = yy++; x4 = xx++; v4 = vv++;
					*y5 += *v5 * *x5;

					y5 = yy++; x5 = xx++; v5 = vv++;
					*y6 += *v6 * *x6;

					y6 = yy++; x6 = xx++; v6 = vv++;
					*y7 += *v7 * *x7;
				}
				*y1 += *v1 * *x1;
				*y2 += *v2 * *x2;
				*y3 += *v3 * *x3;
				*y4 += *v4 * *x4;
				*y5 += *v5 * *x5;
				*y6 += *v6 * *x6;
				i-=6;
			}
			for(;i<je;i++)
			{ 
				*yy += *vv * *xx;
				yy++; xx++; vv++;
			}
		}
	}
}

void lis_matvec_dia_p8s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	LIS_SCALAR *vv,*xx,*yy;
	LIS_SCALAR *y1,*y2,*y3,*y4,*y5,*y6,*y7,*y8;
	LIS_SCALAR *x1,*x2,*x3,*x4,*x5,*x6,*x7,*x8;
	LIS_SCALAR *v1,*v2,*v3,*v4,*v5,*v6,*v7,*v8;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			kk = j*n;
			yy = &y[js];
			xx = &x[jj+js];
			vv = &A->value[kk+js];

			if( je-js<9 )
			{
				i = js;
			}
			else
			{
				y1 = yy++; x1 = xx++; v1 = vv++;
				y2 = yy++; x2 = xx++; v2 = vv++;
				y3 = yy++; x3 = xx++; v3 = vv++;
				y4 = yy++; x4 = xx++; v4 = vv++;
				y5 = yy++; x5 = xx++; v5 = vv++;
				y6 = yy++; x6 = xx++; v6 = vv++;
				y7 = yy++; x7 = xx++; v7 = vv++;
				for(i=js+14;i<je;i+=8)
				{
					y8 = yy++; x8 = xx++; v8 = vv++;
					*y1 += *v1 * *x1;

					y1 = yy++; x1 = xx++; v1 = vv++;
					*y2 += *v2 * *x2;

					y2 = yy++; x2 = xx++; v2 = vv++;
					*y3 += *v3 * *x3;

					y3 = yy++; x3 = xx++; v3 = vv++;
					*y4 += *v4 * *x4;

					y4 = yy++; x4 = xx++; v4 = vv++;
					*y5 += *v5 * *x5;

					y5 = yy++; x5 = xx++; v5 = vv++;
					*y6 += *v6 * *x6;

					y6 = yy++; x6 = xx++; v6 = vv++;
					*y7 += *v7 * *x7;

					y7 = yy++; x7 = xx++; v7 = vv++;
					*y8 += *v8 * *x8;
				}
				*y1 += *v1 * *x1;
				*y2 += *v2 * *x2;
				*y3 += *v3 * *x3;
				*y4 += *v4 * *x4;
				*y5 += *v5 * *x5;
				*y6 += *v6 * *x6;
				*y7 += *v7 * *x7;
				i-=7;
			}
			for(;i<je;i++)
			{ 
				*yy += *vv * *xx;
				yy++; xx++; vv++;
			}
		}
	}
}

