#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#include "lis.h"
#include "lis_io.h"

char *mattypename[] = {	"   ", "crs", "ccs", "msr",
						"dia", "ell", "jds", "bsr", "bsc",
						"vbr","coo", "dns"
						};
char *fileformat_name[] = {"mm", "hb", "lis", "free"};
char *arrayorder_name[] = {"row", "col", "ptr", "index", "bptr",
							"bindex", "value", "solution", "rhs", "diag",
							"lrow", "lcol", "lindex", "lvalue", "urow",
							"ucol", "uindex", "uvalue"
};
char *paramorder_name[] = {"n", "nnz", "ndz", "nnd", "maxnzr", 
							"nr", "nc", "bnnz", "bnr", "bnc", "nnzl",
							"nnzu"
};



int lis_output(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, char *path)
{
	int				err;
	FILE			*file;

	if( A==NULL )
	{
		SETERR("matrix is NULL\n");
		return LIS_FAILS;
	}
	if( A->matrix_type<=0 )
	{
		SETERR("matrix is undefined\n");
		return LIS_FAILS;
	}

	if( format==LIS_FMT_MM )
	{
		switch( A->matrix_type )
		{
		case LIS_MATRIX_CRS:
			err = lis_output_mm_crs(A,b,x,path);
			break;
		case LIS_MATRIX_CCS:
			err = lis_output_mm_ccs(A,b,x,path);
			break;
		case LIS_MATRIX_MSR:
			err = lis_output_mm_msr(A,b,x,path);
			break;
		case LIS_MATRIX_DIA:
			err = lis_output_mm_dia(A,b,x,path);
			break;
		case LIS_MATRIX_ELL:
			err = lis_output_mm_ell(A,b,x,path);
			break;
		case LIS_MATRIX_JDS:
			err = lis_output_mm_jds(A,b,x,path);
			break;
		case LIS_MATRIX_BSR:
			err = lis_output_mm_bsr(A,b,x,path);
			break;
		case LIS_MATRIX_BSC:
			err = lis_output_mm_bsc(A,b,x,path);
			break;
		case LIS_MATRIX_VBR:
			err = lis_output_mm_vbr(A,b,x,path);
			break;
		case LIS_MATRIX_DNS:
			err = lis_output_mm_dns(A,b,x,path);
			break;
		case LIS_MATRIX_COO:
			err = lis_output_mm_coo(A,b,x,path);
			break;
		}
	}
	else
	{
		switch( A->matrix_type )
		{
		case LIS_MATRIX_CRS:
			err = lis_output_lis_crs(A,b,x,format,path);
			break;
		case LIS_MATRIX_CCS:
			err = lis_output_lis_ccs(A,b,x,format,path);
			break;
		case LIS_MATRIX_MSR:
			err = lis_output_lis_msr(A,b,x,format,path);
			break;
		case LIS_MATRIX_DIA:
			err = lis_output_lis_dia(A,b,x,format,path);
			break;
		case LIS_MATRIX_ELL:
			err = lis_output_lis_ell(A,b,x,format,path);
			break;
		case LIS_MATRIX_JDS:
			err = lis_output_lis_jds(A,b,x,format,path);
			break;
		case LIS_MATRIX_BSR:
			err = lis_output_lis_bsr(A,b,x,format,path);
			break;
		case LIS_MATRIX_BSC:
			err = lis_output_lis_bsc(A,b,x,format,path);
			break;
		case LIS_MATRIX_VBR:
			err = lis_output_lis_vbr(A,b,x,format,path);
			break;
		case LIS_MATRIX_DNS:
			err = lis_output_lis_dns(A,b,x,format,path);
			break;
		case LIS_MATRIX_COO:
			err = lis_output_lis_coo(A,b,x,format,path);
			break;
		}
	}
	return err;
}

int lis_output_lis_crs(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, char *path)
{
	int				n,nprocs,my_rank;
	int				i,pe;
	int				err;
	int				count;
	int				*ibuf;
	char			cbuf[128];
	char			buf[128];
	size_t			*tbuf,size;
	FILE			*file;


	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	n         = A->n;

	ibuf = (int *)buf;
	ibuf[0] = ibuf[1] = 0;
	if( b!=NULL ) ibuf[0] = 1;
	if( x!=NULL ) ibuf[1] = 1;
	ibuf[2] = A->nprocs;
	if( format==LIS_FMT_LIS_ASCII )
	{
		if( my_rank==0 )
		{
			file = fopen(path, "w");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}

			fprintf(file, "#LIS A %s\n", mattypename[A->matrix_type]);
			fprintf(file, "%d %d %d\n", ibuf[0], ibuf[1], ibuf[2]);
			fclose(file);
		}

		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "a");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				fprintf(file, "# %d %d %d\n", pe, A->n, A->nnz);

				// ptr
				for(i=0;i<=A->n;i++)
				{
					fprintf(file, "%10d ",A->ptr[i]);
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( (A->n+1)%7!=0 ) fprintf(file, "\n");
				// index
				for(i=0;i<A->nnz;i++)
				{
					#ifdef _MPI
						fprintf(file, "%10d ",A->g_index[i]);
					#else
						fprintf(file, "%10d ",A->index[i]);
					#endif
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( A->nnz%7!=0 ) fprintf(file, "\n");
				// value
				for(i=0;i<A->nnz;i++)
				{
					fprintf(file, "%28.20e ",A->value[i]);
					if( (i+1)%3==0 ) fprintf(file, "\n");
				}
				if( A->nnz%3!=0 ) fprintf(file, "\n");
				// RHS
				if( b!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",b->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				// solution
				if( x!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",x->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}

	}
	else
	{
		format = 1;
		format = *(char *)&format;
		tbuf   = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		// pointer
		tbuf[0]  = 10 + 3*sizeof(int) + (nprocs+1)*sizeof(size_t);
		size     = 2*sizeof(int);
		size    += (A->n+1)*sizeof(int) + A->nnz*sizeof(int) + A->nnz*sizeof(LIS_SCALAR);
		if( b!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		if( x!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		#ifdef _MPI
		MPI_Gather(&size,sizeof(size_t)/sizeof(int),MPI_INT,&tbuf[1],sizeof(size_t)/sizeof(int),MPI_INT,0,A->comm);
		for(i=0;i<nprocs;i++) tbuf[i+1] += tbuf[i];
		#else
		tbuf[1] = size;
		#endif
		if( my_rank==0 )
		{
			file = fopen(path, "wb");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			if( format==LIS_BINARY_BIG )
			{
				sprintf(cbuf, "#LIS B %s", mattypename[A->matrix_type]);
			}
			else
			{
				sprintf(cbuf, "#LIS L %s", mattypename[A->matrix_type]);
			}
			fwrite(cbuf, sizeof(char), 10, file);
			fwrite(ibuf, sizeof(int), 3, file);
			fwrite(tbuf, sizeof(size_t), nprocs+1, file);
			fclose(file);
		}
		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "ab");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				ibuf[0] = A->n;
				ibuf[1] = A->nnz;
				count = fwrite(ibuf, sizeof(int), 2, file);
				// ptr
				count = fwrite(A->ptr, sizeof(int), A->n+1, file);
				// index
				#ifdef _MPI
					count = fwrite(A->g_index, sizeof(int), A->nnz, file);
				#else
					count = fwrite(A->index, sizeof(int), A->nnz, file);
				#endif
				// value
				count = fwrite(A->value, sizeof(LIS_SCALAR), A->nnz, file);
				// RHS
				if( b!=NULL ) count = fwrite(b->value, sizeof(LIS_SCALAR), A->n, file);
				// solution
				if( x!=NULL ) count = fwrite(x->value, sizeof(LIS_SCALAR), A->n, file);
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}
		lis_free(tbuf);
	}
	return LIS_SUCCESS;
}

int lis_output_lis_ccs(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, char *path)
{
	int				n,nprocs,my_rank;
	int				i,pe;
	int				err;
	int				count;
	int				*ibuf;
	char			cbuf[128];
	char			buf[128];
	size_t			*tbuf,size;
	FILE			*file;


	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	n         = A->n;

	ibuf = (int *)buf;
	ibuf[0] = ibuf[1] = 0;
	if( b!=NULL ) ibuf[0] = 1;
	if( x!=NULL ) ibuf[1] = 1;
	ibuf[2] = A->nprocs;
	if( format==LIS_FMT_LIS_ASCII )
	{
		if( my_rank==0 )
		{
			file = fopen(path, "w");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}

			fprintf(file, "#LIS A %s\n", mattypename[A->matrix_type]);
			fprintf(file, "%d %d %d\n", ibuf[0], ibuf[1], ibuf[2]);
			fclose(file);
		}

		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "a");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				fprintf(file, "# %d %d %d\n", pe, A->n, A->nnz);

				// ptr
				for(i=0;i<=A->n;i++)
				{
					fprintf(file, "%10d ",A->ptr[i]);
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( (A->n+1)%7!=0 ) fprintf(file, "\n");
				// index
				for(i=0;i<A->nnz;i++)
				{
					#ifdef _MPI
						fprintf(file, "%10d ",A->g_index[i]);
					#else
						fprintf(file, "%10d ",A->index[i]);
					#endif
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( A->nnz%7!=0 ) fprintf(file, "\n");
				// value
				for(i=0;i<A->nnz;i++)
				{
					fprintf(file, "%28.20e ",A->value[i]);
					if( (i+1)%3==0 ) fprintf(file, "\n");
				}
				if( A->nnz%3!=0 ) fprintf(file, "\n");
				// RHS
				if( b!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",b->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				// solution
				if( x!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",x->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}

	}
	else
	{
		format = 1;
		format = *(char *)&format;
		// pointer
		tbuf   = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		tbuf[0]  = 10 + 3*sizeof(int) + (nprocs+1)*sizeof(size_t);
		size     = 2*sizeof(int);
		size    += (A->n+1)*sizeof(int) + A->nnz*sizeof(int) + A->nnz*sizeof(LIS_SCALAR);
		if( b!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		if( x!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		#ifdef _MPI
		MPI_Gather(&size,sizeof(size_t)/sizeof(int),MPI_INT,&tbuf[1],sizeof(size_t)/sizeof(int),MPI_INT,0,A->comm);
		for(i=0;i<nprocs;i++) tbuf[i+1] += tbuf[i];
		#else
		tbuf[1] = size;
		#endif
		if( my_rank==0 )
		{
			file = fopen(path, "wb");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			if( format==LIS_BINARY_BIG )
			{
				sprintf(cbuf, "#LIS B %s", mattypename[A->matrix_type]);
			}
			else
			{
				sprintf(cbuf, "#LIS L %s", mattypename[A->matrix_type]);
			}
			fwrite(cbuf, sizeof(char), 10, file);
			fwrite(ibuf, sizeof(int), 3, file);
			fwrite(tbuf, sizeof(size_t), nprocs+1, file);
			fclose(file);
		}
		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "ab");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				ibuf[0] = A->n;
				ibuf[1] = A->nnz;
				count = fwrite(ibuf, sizeof(int), 2, file);
				// ptr
				count = fwrite(A->ptr, sizeof(int), A->n+1, file);
				// index
				#ifdef _MPI
					count = fwrite(A->g_index, sizeof(int), A->nnz, file);
				#else
					count = fwrite(A->index, sizeof(int), A->nnz, file);
				#endif
				// value
				count = fwrite(A->value, sizeof(LIS_SCALAR), A->nnz, file);
				// RHS
				if( b!=NULL ) count = fwrite(b->value, sizeof(LIS_SCALAR), A->n, file);
				// solution
				if( x!=NULL ) count = fwrite(x->value, sizeof(LIS_SCALAR), A->n, file);
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}
		lis_free(tbuf);
	}
	return LIS_SUCCESS;
}

int lis_output_lis_msr(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, char *path)
{
	int				n,nprocs,my_rank;
	int				i,pe;
	int				err;
	int				count;
	int				*ibuf;
	char			cbuf[128];
	char			buf[128];
	size_t			*tbuf,size;
	FILE			*file;


	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	n         = A->n;

	ibuf = (int *)buf;
	ibuf[0] = ibuf[1] = 0;
	if( b!=NULL ) ibuf[0] = 1;
	if( x!=NULL ) ibuf[1] = 1;
	ibuf[2] = A->nprocs;
	if( format==LIS_FMT_LIS_ASCII )
	{
		if( my_rank==0 )
		{
			file = fopen(path, "w");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}

			fprintf(file, "#LIS A %s\n", mattypename[A->matrix_type]);
			fprintf(file, "%d %d %d\n", ibuf[0], ibuf[1], ibuf[2]);
			fclose(file);
		}

		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "a");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				fprintf(file, "# %d %d %d %d\n", pe, A->n, A->nnz, A->ndz);

				// index
				for(i=0;i<A->nnz+A->ndz+1;i++)
				{
					#ifdef _MPI
						fprintf(file, "%10d ",A->g_index[i]);
					#else
						fprintf(file, "%10d ",A->index[i]);
					#endif
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( (A->nnz+A->ndz+1)%7!=0 ) fprintf(file, "\n");
				// value
				for(i=0;i<A->nnz+A->ndz+1;i++)
				{
					fprintf(file, "%28.20e ",A->value[i]);
					if( (i+1)%3==0 ) fprintf(file, "\n");
				}
				if( (A->nnz+A->ndz+1)%3!=0 ) fprintf(file, "\n");
				// RHS
				if( b!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",b->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				// solution
				if( x!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",x->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}

	}
	else
	{
		format = 1;
		format = *(char *)&format;
		// pointer
		tbuf   = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		tbuf[0]  = 10 + 3*sizeof(int) + (nprocs+1)*sizeof(size_t);
		size     = 3*sizeof(int);
		size    += (A->nnz+A->ndz+1)*(sizeof(int) + sizeof(LIS_SCALAR));
		if( b!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		if( x!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		#ifdef _MPI
		MPI_Gather(&size,sizeof(size_t)/sizeof(int),MPI_INT,&tbuf[1],sizeof(size_t)/sizeof(int),MPI_INT,0,A->comm);
		for(i=0;i<nprocs;i++) tbuf[i+1] += tbuf[i];
		#else
		tbuf[1] = size;
		#endif
		if( my_rank==0 )
		{
			file = fopen(path, "wb");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			if( format==LIS_BINARY_BIG )
			{
				sprintf(cbuf, "#LIS B %s", mattypename[A->matrix_type]);
			}
			else
			{
				sprintf(cbuf, "#LIS L %s", mattypename[A->matrix_type]);
			}
			fwrite(cbuf, sizeof(char), 10, file);
			fwrite(ibuf, sizeof(int), 3, file);
			fwrite(tbuf, sizeof(size_t), nprocs+1, file);
			fclose(file);
		}
		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "ab");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				ibuf[0] = A->n;
				ibuf[1] = A->nnz;
				ibuf[2] = A->ndz;
				count = fwrite(ibuf, sizeof(int), 3, file);
				// index
				#ifdef _MPI
					count = fwrite(A->g_index, sizeof(int), A->nnz+A->ndz+1, file);
				#else
					count = fwrite(A->index, sizeof(int), A->nnz+A->ndz+1, file);
				#endif
				// value
				count = fwrite(A->value, sizeof(LIS_SCALAR), A->nnz+A->ndz+1, file);
				// RHS
				if( b!=NULL ) count = fwrite(b->value, sizeof(LIS_SCALAR), A->n, file);
				// solution
				if( x!=NULL ) count = fwrite(x->value, sizeof(LIS_SCALAR), A->n, file);
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}
		lis_free(tbuf);
	}
	return LIS_SUCCESS;
}

int lis_output_lis_dia(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, char *path)
{
#ifdef _OPENMP
	int				n,nprocs,my_rank;
	int				i,j,k,pe;
	int				err;
	int				count;
	int				*ibuf;
	char			cbuf[128];
	char			buf[128];
	size_t			*tbuf,size;
	FILE			*file;


	nprocs    = omp_get_max_threads();
	my_rank   = A->my_rank;
	n         = A->n;

	ibuf = (int *)buf;
	ibuf[0] = ibuf[1] = 0;
	if( b!=NULL ) ibuf[0] = 1;
	if( x!=NULL ) ibuf[1] = 1;
	ibuf[2] = nprocs;
	if( format==LIS_FMT_LIS_ASCII )
	{
		if( my_rank==0 )
		{
			file = fopen(path, "w");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}

			fprintf(file, "#LIS A %s\n", mattypename[A->matrix_type]);
			fprintf(file, "%d %d %d\n", ibuf[0], ibuf[1], ibuf[2]);
			fclose(file);
		}
		pe = 0;
		{
			{
				file = fopen(path, "a");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				fprintf(file, "# %d %d %d\n", pe, A->n, A->nnd);

				// nnd
				for(i=0;i<nprocs;i++)
				{
					fprintf(file, "%10d ",A->col[i]);
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( nprocs%7!=0 ) fprintf(file, "\n");
				// index
				k = 0;
				for(j=0;j<nprocs;j++)
				{
					for(i=0;i<A->nnd;i++)
					{
						fprintf(file, "%10d ",A->index[j*A->nnd + i]);
						if( (k+1)%7==0 ) fprintf(file, "\n");
						k++;
					}
				}
				if( k%7!=0 ) fprintf(file, "\n");
				// value
				for(i=0;i<A->n*A->nnd;i++)
				{
					fprintf(file, "%28.20e ",A->value[i]);
					if( (i+1)%3==0 ) fprintf(file, "\n");
				}
				if( A->nnd*A->n%3!=0 ) fprintf(file, "\n");
				// RHS
				if( b!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",b->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				// solution
				if( x!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",x->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				fclose(file);
			}
		}

	}
	else
	{
		format = 1;
		format = *(char *)&format;
		// pointer
		tbuf   = (size_t *)lis_malloc( 2*sizeof(size_t) );
		tbuf[0]  = 10 + 3*sizeof(int) + 2*sizeof(size_t);
		size     = 2*sizeof(int);
		size    += nprocs*sizeof(int) + nprocs*A->nnd*sizeof(int) + A->nnd*A->n*sizeof(LIS_SCALAR);
		if( b!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		if( x!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		tbuf[1] = size;
		if( my_rank==0 )
		{
			file = fopen(path, "wb");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			if( format==LIS_BINARY_BIG )
			{
				sprintf(cbuf, "#LIS B %s", mattypename[A->matrix_type]);
			}
			else
			{
				sprintf(cbuf, "#LIS L %s", mattypename[A->matrix_type]);
			}
			fwrite(cbuf, sizeof(char), 10, file);
			fwrite(ibuf, sizeof(int), 3, file);
			fwrite(tbuf, sizeof(size_t), 2, file);
			fclose(file);
		}
		{
			if( my_rank==pe )
			{
				file = fopen(path, "ab");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				ibuf[0] = A->n;
				ibuf[1] = A->nnd;
				count = fwrite(ibuf, sizeof(int), 2, file);
				// nnd
				count = fwrite(A->col, sizeof(int), nprocs, file);
				// index
				count = fwrite(A->index, sizeof(int), nprocs*A->nnd, file);
				// value
				count = fwrite(A->value, sizeof(LIS_SCALAR), A->nnd*A->n, file);
				// RHS
				if( b!=NULL ) count = fwrite(b->value, sizeof(LIS_SCALAR), A->n, file);
				// solution
				if( x!=NULL ) count = fwrite(x->value, sizeof(LIS_SCALAR), A->n, file);
				fclose(file);
			}
		}
		lis_free(tbuf);
	}
	return LIS_SUCCESS;
#else
#ifdef _MPI
	int				gn,n,nnd,nprocs,my_rank;
	int				i,j,jj,js,je,is,pe;
	int				err;
	int				count;
	int				*ibuf;
	int				*iw,*index;
	char			cbuf[128];
	char			buf[128];
	size_t			*tbuf,size;
	FILE			*file;
	LIS_SCALAR		*value;


	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	n         = A->n;
	gn        = A->gn;
	is        = A->is;

	index     = NULL;
	value     = NULL;
	iw        = NULL;


	iw = (int *)lis_malloc( (n+gn)*sizeof(int) );
	if( iw==NULL )
	{
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	for(i=0;i<n+gn;i++) iw[i] = 0;
	nnd = 0;
	for(j=0;j<A->nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,gn-jj);
		for(i=js;i<je;i++)
		{
			if( jj+i<A->np && A->value[j*n+i]!=0.0 )
			{
				iw[n+A->g_index[jj+i]-i] = 1;
			}
		}
	}
	for(i=0;i<(n+gn);i++)
	{
		if( iw[i] > 0 ) nnd++;
	}
	index = (int *)lis_malloc( nnd*sizeof(int) );
	if( index==NULL )
	{
		lis_free2(3,iw,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	value = (LIS_SCALAR *)lis_malloc( n*nnd*sizeof(LIS_SCALAR) );
	if( value==NULL )
	{
		lis_free2(3,iw,index,value);
		SETERR("out of memory\n");
		return LIS_FAILS;
	}
	j = 0;
	for( i=0; i<n+gn && j<nnd; i++ )
	{
		if( iw[i]>0 )
		{
			index[j]   = i-is-n;
			iw[i]  = j*n;
			j++;
		}
	}
	memset(value,0,n*nnd*sizeof(LIS_SCALAR));
	for(j=0;j<A->nnd;j++)
	{
		jj = A->index[j];
		js = _max(0,-jj);
		je = _min(n,gn-jj);
		for(i=js;i<je;i++)
		{
			if( jj+i<A->np && A->value[j*n+i]!=0.0 )
			{
				value[iw[n+A->g_index[jj+i]-i] + i] = A->value[j*n+i];
			}
		}
	}

	ibuf = (int *)buf;
	ibuf[0] = ibuf[1] = 0;
	if( b!=NULL ) ibuf[0] = 1;
	if( x!=NULL ) ibuf[1] = 1;
	ibuf[2] = A->nprocs;
	if( format==LIS_FMT_LIS_ASCII )
	{
		if( my_rank==0 )
		{
			file = fopen(path, "w");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				lis_free2(3,iw,index,value);
				return LIS_FAILS;
			}

			fprintf(file, "#LIS A %s\n", mattypename[A->matrix_type]);
			fprintf(file, "%d %d %d\n", ibuf[0], ibuf[1], ibuf[2]);
			fclose(file);
		}

		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "a");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					lis_free2(3,iw,index,value);
					return LIS_FAILS;
				}
				// params
				fprintf(file, "# %d %d %d\n", pe, A->n, nnd);

				// index
				for(i=0;i<nnd;i++)
				{
					fprintf(file, "%10d ",index[i]);
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( nnd%7!=0 ) fprintf(file, "\n");
				// value
				for(i=0;i<n*nnd;i++)
				{
					fprintf(file, "%28.20e ",value[i]);
					if( (i+1)%3==0 ) fprintf(file, "\n");
				}
				if( nnd*n%3!=0 ) fprintf(file, "\n");
				// RHS
				if( b!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",b->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				// solution
				if( x!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",x->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}

	}
	else
	{
		format = 1;
		format = *(char *)&format;
		// pointer
		tbuf   = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		tbuf[0]  = 10 + 3*sizeof(int) + (nprocs+1)*sizeof(size_t);
		size     = 2*sizeof(int);
		size    += nnd*sizeof(int) + nnd*n*sizeof(LIS_SCALAR);
		if( b!=NULL ) size += n*sizeof(LIS_SCALAR);
		if( x!=NULL ) size += n*sizeof(LIS_SCALAR);
		MPI_Gather(&size,sizeof(size_t)/sizeof(int),MPI_INT,&tbuf[1],sizeof(size_t)/sizeof(int),MPI_INT,0,A->comm);
		for(i=0;i<nprocs;i++) tbuf[i+1] += tbuf[i];
		if( my_rank==0 )
		{
			file = fopen(path, "wb");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				lis_free2(3,iw,index,value);
				return LIS_FAILS;
			}
			if( format==LIS_BINARY_BIG )
			{
				sprintf(cbuf, "#LIS B %s", mattypename[A->matrix_type]);
			}
			else
			{
				sprintf(cbuf, "#LIS L %s", mattypename[A->matrix_type]);
			}
			fwrite(cbuf, sizeof(char), 10, file);
			fwrite(ibuf, sizeof(int), 3, file);
			fwrite(tbuf, sizeof(size_t), nprocs+1, file);
			fclose(file);
		}
		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "ab");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					lis_free2(3,iw,index,value);
					return LIS_FAILS;
				}
				// params
				ibuf[0] = n;
				ibuf[1] = nnd;
				count = fwrite(ibuf, sizeof(int), 2, file);
				// index
				count = fwrite(index, sizeof(int), nnd, file);
				// value
				count = fwrite(value, sizeof(LIS_SCALAR), nnd*n, file);
				// RHS
				if( b!=NULL ) count = fwrite(b->value, sizeof(LIS_SCALAR), A->n, file);
				// solution
				if( x!=NULL ) count = fwrite(x->value, sizeof(LIS_SCALAR), A->n, file);
				fclose(file);
			}
			MPI_Barrier(A->comm);
		}
		lis_free(tbuf);
	}
	lis_free2(3,iw,index,value);
	return LIS_SUCCESS;
#else
	int				gn,n,nnd,nprocs,my_rank;
	int				i,j,jj,js,je,is,pe;
	int				err;
	int				count;
	int				*ibuf;
	char			cbuf[128];
	char			buf[128];
	size_t			*tbuf,size;
	FILE			*file;


	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	n         = A->n;
	gn        = A->gn;
	is        = A->is;

	ibuf = (int *)buf;
	ibuf[0] = ibuf[1] = 0;
	if( b!=NULL ) ibuf[0] = 1;
	if( x!=NULL ) ibuf[1] = 1;
	ibuf[2] = A->nprocs;
	if( format==LIS_FMT_LIS_ASCII )
	{
		if( my_rank==0 )
		{
			file = fopen(path, "w");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}

			fprintf(file, "#LIS A %s\n", mattypename[A->matrix_type]);
			fprintf(file, "%d %d %d\n", ibuf[0], ibuf[1], ibuf[2]);
			fclose(file);
		}

		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "a");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				fprintf(file, "# %d %d %d\n", pe, A->n, A->nnd);

				// index
				for(i=0;i<A->nnd;i++)
				{
					fprintf(file, "%10d ",A->index[i]);
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( A->nnd%7!=0 ) fprintf(file, "\n");
				// value
				for(i=0;i<A->n*A->nnd;i++)
				{
					fprintf(file, "%28.20e ",A->value[i]);
					if( (i+1)%3==0 ) fprintf(file, "\n");
				}
				if( A->nnd*A->n%3!=0 ) fprintf(file, "\n");
				// RHS
				if( b!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",b->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				// solution
				if( x!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",x->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				fclose(file);
			}
		}

	}
	else
	{
		format = 1;
		format = *(char *)&format;
		// pointer
		tbuf   = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		tbuf[0]  = 10 + 3*sizeof(int) + (nprocs+1)*sizeof(size_t);
		size     = 2*sizeof(int);
		size    += A->nnd*sizeof(int) + A->nnd*A->n*sizeof(LIS_SCALAR);
		if( b!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		if( x!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		tbuf[1] = size;
		if( my_rank==0 )
		{
			file = fopen(path, "wb");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			if( format==LIS_BINARY_BIG )
			{
				sprintf(cbuf, "#LIS B %s", mattypename[A->matrix_type]);
			}
			else
			{
				sprintf(cbuf, "#LIS L %s", mattypename[A->matrix_type]);
			}
			fwrite(cbuf, sizeof(char), 10, file);
			fwrite(ibuf, sizeof(int), 3, file);
			fwrite(tbuf, sizeof(size_t), nprocs+1, file);
			fclose(file);
		}
		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "ab");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				ibuf[0] = A->n;
				ibuf[1] = A->nnd;
				count = fwrite(ibuf, sizeof(int), 2, file);
				// index
				count = fwrite(A->index, sizeof(int), A->nnd, file);
				// value
				count = fwrite(A->value, sizeof(LIS_SCALAR), A->nnd*A->n, file);
				// RHS
				if( b!=NULL ) count = fwrite(b->value, sizeof(LIS_SCALAR), A->n, file);
				// solution
				if( x!=NULL ) count = fwrite(x->value, sizeof(LIS_SCALAR), A->n, file);
				fclose(file);
			}
		}
		lis_free(tbuf);
	}
	return LIS_SUCCESS;
#endif
#endif
}

int lis_output_lis_ell(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, char *path)
{
	int				n,nprocs,my_rank;
	int				i,pe;
	int				err;
	int				count;
	int				*ibuf;
	char			cbuf[128];
	char			buf[128];
	size_t			*tbuf,size;
	FILE			*file;


	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	n         = A->n;

	ibuf = (int *)buf;
	ibuf[0] = ibuf[1] = 0;
	if( b!=NULL ) ibuf[0] = 1;
	if( x!=NULL ) ibuf[1] = 1;
	ibuf[2] = A->nprocs;
	if( format==LIS_FMT_LIS_ASCII )
	{
		if( my_rank==0 )
		{
			file = fopen(path, "w");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}

			fprintf(file, "#LIS A %s\n", mattypename[A->matrix_type]);
			fprintf(file, "%d %d %d\n", ibuf[0], ibuf[1], ibuf[2]);
			fclose(file);
		}

		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "a");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				fprintf(file, "# %d %d %d\n", pe, A->n, A->maxnzr);

				// index
				for(i=0;i<A->n*A->maxnzr;i++)
				{
					#ifdef _MPI
						fprintf(file, "%10d ",A->g_index[i]);
					#else
						fprintf(file, "%10d ",A->index[i]);
					#endif
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( A->n*A->maxnzr%7!=0 ) fprintf(file, "\n");
				// value
				for(i=0;i<A->n*A->maxnzr;i++)
				{
					fprintf(file, "%28.20e ",A->value[i]);
					if( (i+1)%3==0 ) fprintf(file, "\n");
				}
				if( A->n*A->maxnzr%3!=0 ) fprintf(file, "\n");
				// RHS
				if( b!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",b->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				// solution
				if( x!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",x->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}

	}
	else
	{
		format = 1;
		format = *(char *)&format;
		// pointer
		tbuf   = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		tbuf[0]  = 10 + 3*sizeof(int) + (nprocs+1)*sizeof(size_t);
		size     = 2*sizeof(int);
		size    += A->n*A->maxnzr*(sizeof(int) + sizeof(LIS_SCALAR));
		if( b!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		if( x!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		#ifdef _MPI
		MPI_Gather(&size,sizeof(size_t)/sizeof(int),MPI_INT,&tbuf[1],sizeof(size_t)/sizeof(int),MPI_INT,0,A->comm);
		for(i=0;i<nprocs;i++) tbuf[i+1] += tbuf[i];
		#else
		tbuf[1] = size;
		#endif
		if( my_rank==0 )
		{
			file = fopen(path, "wb");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			if( format==LIS_BINARY_BIG )
			{
				sprintf(cbuf, "#LIS B %s", mattypename[A->matrix_type]);
			}
			else
			{
				sprintf(cbuf, "#LIS L %s", mattypename[A->matrix_type]);
			}
			fwrite(cbuf, sizeof(char), 10, file);
			fwrite(ibuf, sizeof(int), 3, file);
			fwrite(tbuf, sizeof(size_t), nprocs+1, file);
			fclose(file);
		}
		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "ab");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				ibuf[0] = A->n;
				ibuf[1] = A->maxnzr;
				count = fwrite(ibuf, sizeof(int), 2, file);
				// index
				#ifdef _MPI
					count = fwrite(A->g_index, sizeof(int), A->n*A->maxnzr, file);
				#else
					count = fwrite(A->index, sizeof(int), A->n*A->maxnzr, file);
				#endif
				// value
				count = fwrite(A->value, sizeof(LIS_SCALAR), A->n*A->maxnzr, file);
				// RHS
				if( b!=NULL ) count = fwrite(b->value, sizeof(LIS_SCALAR), A->n, file);
				// solution
				if( x!=NULL ) count = fwrite(x->value, sizeof(LIS_SCALAR), A->n, file);
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}
		lis_free(tbuf);
	}
	return LIS_SUCCESS;
}

int lis_output_lis_jds(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, char *path)
{
#ifdef _OPENMP
	int				n,nprocs,my_rank,maxmaxnzr;
	int				i,pe,is,ie,j,k;
	int				err;
	int				count;
	int				*ibuf;
	char			cbuf[128];
	char			buf[128];
	size_t			*tbuf,size;
	FILE			*file;


	nprocs    = omp_get_max_threads();
	my_rank   = A->my_rank;
	n         = A->n;
	maxmaxnzr = A->maxnzr;


	ibuf = (int *)buf;
	ibuf[0] = ibuf[1] = 0;
	if( b!=NULL ) ibuf[0] = 1;
	if( x!=NULL ) ibuf[1] = 1;
	ibuf[2] = nprocs;
	if( format==LIS_FMT_LIS_ASCII )
	{
		if( my_rank==0 )
		{
			file = fopen(path, "w");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}

			fprintf(file, "#LIS A %s\n", mattypename[A->matrix_type]);
			fprintf(file, "%d %d %d\n", ibuf[0], ibuf[1], ibuf[2]);
			fclose(file);
		}

			file = fopen(path, "a");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			// params
			pe = 0;
			fprintf(file, "# %d %d %d %d\n", pe, A->n, A->nnz, A->maxnzr);
		{
				// maxnzr
				for(i=0;i<nprocs;i++)
				{
					fprintf(file, "%10d ",A->col[i]);
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( nprocs%7!=0 ) fprintf(file, "\n");
				// row
				for(i=0;i<n;i++)
				{
					fprintf(file, "%10d ",A->row[i]);
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( A->n%7!=0 ) fprintf(file, "\n");
				// ptr
				k = 0;
				for(j=0;j<nprocs;j++)
				{
					for(i=0;i<maxmaxnzr+1;i++)
					{
						fprintf(file, "%10d ",A->ptr[j*(maxmaxnzr+1) + i]);
						if( (k+1)%7==0 ) fprintf(file, "\n");
						k++;
					}
				}
				if( k%7!=0 ) fprintf(file, "\n");
				// index
				for(i=0;i<A->nnz;i++)
				{
					#ifdef _MPI
						fprintf(file, "%10d ",A->g_index[i]);
					#else
						fprintf(file, "%10d ",A->index[i]);
					#endif
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( A->nnz%7!=0 ) fprintf(file, "\n");
				// value
				for(i=0;i<A->nnz;i++)
				{
					fprintf(file, "%28.20e ",A->value[i]);
					if( (i+1)%3==0 ) fprintf(file, "\n");
				}
				if( A->nnz%3!=0 ) fprintf(file, "\n");
				// RHS
				if( b!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",b->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				// solution
				if( x!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",x->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}

			fclose(file);
		}

	}
	else
	{
		format = 1;
		format = *(char *)&format;
		// pointer
		tbuf   = (size_t *)lis_malloc( 2*sizeof(size_t) );
		tbuf[0]  = 10 + 3*sizeof(int) + 2*sizeof(size_t);
		size     = 3*sizeof(int);
		size    += (nprocs+A->n+nprocs*(A->maxnzr+1)+A->nnz)*sizeof(int) + A->nnz*sizeof(LIS_SCALAR);
		if( b!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		if( x!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		tbuf[1] = size;
		if( my_rank==0 )
		{
			file = fopen(path, "wb");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			if( format==LIS_BINARY_BIG )
			{
				sprintf(cbuf, "#LIS B %s", mattypename[A->matrix_type]);
			}
			else
			{
				sprintf(cbuf, "#LIS L %s", mattypename[A->matrix_type]);
			}
			fwrite(cbuf, sizeof(char), 10, file);
			fwrite(ibuf, sizeof(int), 3, file);
			fwrite(tbuf, sizeof(size_t), 2, file);
			fclose(file);
		}
		{
			{
				file = fopen(path, "ab");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				ibuf[0] = A->n;
				ibuf[1] = A->nnz;
				ibuf[2] = A->maxnzr;
				count = fwrite(ibuf, sizeof(int), 3, file);
				// maxnzr
				count = fwrite(A->col, sizeof(int), nprocs, file);
				// row
				count = fwrite(A->row, sizeof(int), A->n, file);
				// ptr
				count = fwrite(A->ptr, sizeof(int), nprocs*(A->maxnzr+1), file);
				// index
				count = fwrite(A->index, sizeof(int), A->nnz, file);
				// value
				count = fwrite(A->value, sizeof(LIS_SCALAR), A->nnz, file);
				// RHS
				if( b!=NULL ) count = fwrite(b->value, sizeof(LIS_SCALAR), A->n, file);
				// solution
				if( x!=NULL ) count = fwrite(x->value, sizeof(LIS_SCALAR), A->n, file);
				fclose(file);
			}
		}
		lis_free(tbuf);
	}
	return LIS_SUCCESS;
#else
	int				n,nprocs,my_rank;
	int				i,is,pe;
	int				err;
	int				count;
	int				*ibuf;
	char			cbuf[128];
	char			buf[128];
	size_t			*tbuf,size;
	FILE			*file;


	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	n         = A->n;
	is        = A->is;

	ibuf = (int *)buf;
	ibuf[0] = ibuf[1] = 0;
	if( b!=NULL ) ibuf[0] = 1;
	if( x!=NULL ) ibuf[1] = 1;
	ibuf[2] = A->nprocs;
	if( format==LIS_FMT_LIS_ASCII )
	{
		if( my_rank==0 )
		{
			file = fopen(path, "w");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}

			fprintf(file, "#LIS A %s\n", mattypename[A->matrix_type]);
			fprintf(file, "%d %d %d\n", ibuf[0], ibuf[1], ibuf[2]);
			fclose(file);
		}

		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "a");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				fprintf(file, "# %d %d %d %d\n", pe, A->n, A->nnz, A->maxnzr);

				// row
				for(i=0;i<A->n;i++)
				{
					fprintf(file, "%10d ",A->row[i]+is);
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( A->n%7!=0 ) fprintf(file, "\n");
				// ptr
				for(i=0;i<=A->maxnzr;i++)
				{
					fprintf(file, "%10d ",A->ptr[i]);
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( (A->maxnzr+1)%7!=0 ) fprintf(file, "\n");
				// index
				for(i=0;i<A->nnz;i++)
				{
					#ifdef _MPI
						fprintf(file, "%10d ",A->g_index[i]);
					#else
						fprintf(file, "%10d ",A->index[i]);
					#endif
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( A->nnz%7!=0 ) fprintf(file, "\n");
				// value
				for(i=0;i<A->nnz;i++)
				{
					fprintf(file, "%28.20e ",A->value[i]);
					if( (i+1)%3==0 ) fprintf(file, "\n");
				}
				if( A->nnz%3!=0 ) fprintf(file, "\n");
				// RHS
				if( b!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",b->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				// solution
				if( x!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",x->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}
	}
	else
	{
		format = 1;
		format = *(char *)&format;
		// pointer
		tbuf   = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		tbuf[0]  = 10 + 3*sizeof(int) + (nprocs+1)*sizeof(size_t);
		size     = 3*sizeof(int);
		size    += (A->n+A->maxnzr+1+A->nnz)*sizeof(int) + A->nnz*sizeof(LIS_SCALAR);
		if( b!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		if( x!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		#ifdef _MPI
		MPI_Gather(&size,sizeof(size_t)/sizeof(int),MPI_INT,&tbuf[1],sizeof(size_t)/sizeof(int),MPI_INT,0,A->comm);
		for(i=0;i<nprocs;i++) tbuf[i+1] += tbuf[i];
		#else
		tbuf[1] = size;
		#endif
		if( my_rank==0 )
		{
			file = fopen(path, "wb");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			if( format==LIS_BINARY_BIG )
			{
				sprintf(cbuf, "#LIS B %s", mattypename[A->matrix_type]);
			}
			else
			{
				sprintf(cbuf, "#LIS L %s", mattypename[A->matrix_type]);
			}
			fwrite(cbuf, sizeof(char), 10, file);
			fwrite(ibuf, sizeof(int), 3, file);
			fwrite(tbuf, sizeof(size_t), nprocs+1, file);
			fclose(file);
		}
		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "ab");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				ibuf[0] = A->n;
				ibuf[1] = A->nnz;
				ibuf[2] = A->maxnzr;
				count = fwrite(ibuf, sizeof(int), 3, file);
				// row
				#ifdef _MPI
				for(i=0;i<n;i++) A->row[i] += is;
				#endif
				count = fwrite(A->row, sizeof(int), A->n, file);
				#ifdef _MPI
				for(i=0;i<n;i++) A->row[i] -= is;
				#endif
				// ptr
				count = fwrite(A->ptr, sizeof(int), A->maxnzr+1, file);
				// index
				#ifdef _MPI
					count = fwrite(A->g_index, sizeof(int), A->nnz, file);
				#else
					count = fwrite(A->index, sizeof(int), A->nnz, file);
				#endif
				// value
				count = fwrite(A->value, sizeof(LIS_SCALAR), A->nnz, file);
				// RHS
				if( b!=NULL ) count = fwrite(b->value, sizeof(LIS_SCALAR), A->n, file);
				// solution
				if( x!=NULL ) count = fwrite(x->value, sizeof(LIS_SCALAR), A->n, file);
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}
		lis_free(tbuf);
	}
	return LIS_SUCCESS;
#endif
}

int lis_output_lis_bsr(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, char *path)
{
	int				n,nprocs,my_rank;
	int				i,pe;
	int				err;
	int				count;
	int				*ibuf;
	char			cbuf[128];
	char			buf[128];
	size_t			*tbuf,size;
	FILE			*file;


	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	n         = A->n;

	ibuf = (int *)buf;
	ibuf[0] = ibuf[1] = 0;
	if( b!=NULL ) ibuf[0] = 1;
	if( x!=NULL ) ibuf[1] = 1;
	ibuf[2] = A->nprocs;
	if( format==LIS_FMT_LIS_ASCII )
	{
		if( my_rank==0 )
		{
			file = fopen(path, "w");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}

			fprintf(file, "#LIS A %s\n", mattypename[A->matrix_type]);
			fprintf(file, "%d %d %d\n", ibuf[0], ibuf[1], ibuf[2]);
			fclose(file);
		}

		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "a");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				fprintf(file, "# %d %d %d %d %d %d %d\n", pe, A->n, A->nr, A->nc, A->bnr, A->bnc, A->bnnz);
				// bptr
				for(i=0;i<=A->nr;i++)
				{
					fprintf(file, "%10d ",A->bptr[i]);
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( (A->nr+1)%7!=0 ) fprintf(file, "\n");
				// bindex
				for(i=0;i<A->bnnz;i++)
				{
					#ifdef _MPI
						fprintf(file, "%10d ",A->g_index[i]);
					#else
						fprintf(file, "%10d ",A->bindex[i]);
					#endif
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( A->bnnz%7!=0 ) fprintf(file, "\n");
				// value
				for(i=0;i<A->bnnz*A->bnr*A->bnc;i++)
				{
					fprintf(file, "%28.20e ",A->value[i]);
					if( (i+1)%3==0 ) fprintf(file, "\n");
				}
				if( (A->bnnz*A->bnr*A->bnc)%3!=0 ) fprintf(file, "\n");
				// RHS
				if( b!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",b->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				// solution
				if( x!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",x->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}

	}
	else
	{
		format = 1;
		format = *(char *)&format;
		// pointer
		tbuf   = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		tbuf[0]  = 10 + 3*sizeof(int) + (nprocs+1)*sizeof(size_t);
		size     = 6*sizeof(int);
		size    += (A->nc+1+A->bnnz)*sizeof(int) + A->bnnz*A->bnr*A->bnc*sizeof(LIS_SCALAR);
		if( b!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		if( x!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		#ifdef _MPI
		MPI_Gather(&size,sizeof(size_t)/sizeof(int),MPI_INT,&tbuf[1],sizeof(size_t)/sizeof(int),MPI_INT,0,A->comm);
		for(i=0;i<nprocs;i++) tbuf[i+1] += tbuf[i];
		#else
		tbuf[1] = size;
		#endif
		if( my_rank==0 )
		{
			file = fopen(path, "wb");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			if( format==LIS_BINARY_BIG )
			{
				sprintf(cbuf, "#LIS B %s", mattypename[A->matrix_type]);
			}
			else
			{
				sprintf(cbuf, "#LIS L %s", mattypename[A->matrix_type]);
			}
			fwrite(cbuf, sizeof(char), 10, file);
			fwrite(ibuf, sizeof(int), 3, file);
			fwrite(tbuf, sizeof(size_t), nprocs+1, file);
			fclose(file);
		}
		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "ab");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				ibuf[0] = A->n;
				ibuf[1] = A->nr;
				ibuf[2] = A->nc;
				ibuf[3] = A->bnr;
				ibuf[4] = A->bnc;
				ibuf[5] = A->bnnz;
				count = fwrite(ibuf, sizeof(int), 6, file);
				// bptr
				count = fwrite(A->bptr, sizeof(int), A->nr+1, file);
				// bindex
				#ifdef _MPI
					count = fwrite(A->g_index, sizeof(int), A->bnnz, file);
				#else
					count = fwrite(A->bindex, sizeof(int), A->bnnz, file);
				#endif
				// value
				count = fwrite(A->value, sizeof(LIS_SCALAR), A->bnnz*A->bnr*A->bnc, file);
				// RHS
				if( b!=NULL ) count = fwrite(b->value, sizeof(LIS_SCALAR), A->n, file);
				// solution
				if( x!=NULL ) count = fwrite(x->value, sizeof(LIS_SCALAR), A->n, file);
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}
		lis_free(tbuf);
	}
	return LIS_SUCCESS;
}

int lis_output_lis_bsc(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, char *path)
{
	int				n,nprocs,my_rank;
	int				i,pe;
	int				err;
	int				count;
	int				*ibuf;
	char			cbuf[128];
	char			buf[128];
	size_t			*tbuf,size;
	FILE			*file;


	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	n         = A->n;

	ibuf = (int *)buf;
	ibuf[0] = ibuf[1] = 0;
	if( b!=NULL ) ibuf[0] = 1;
	if( x!=NULL ) ibuf[1] = 1;
	ibuf[2] = A->nprocs;
	if( format==LIS_FMT_LIS_ASCII )
	{
		if( my_rank==0 )
		{
			file = fopen(path, "w");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}

			fprintf(file, "#LIS A %s\n", mattypename[A->matrix_type]);
			fprintf(file, "%d %d %d\n", ibuf[0], ibuf[1], ibuf[2]);
			fclose(file);
		}

		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "a");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				fprintf(file, "# %d %d %d %d %d %d %d\n", pe, A->n, A->nr, A->nc, A->bnr, A->bnc, A->bnnz);
				// bptr
				for(i=0;i<=A->nc;i++)
				{
					fprintf(file, "%10d ",A->bptr[i]);
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( (A->nc+1)%7!=0 ) fprintf(file, "\n");
				// bindex
				for(i=0;i<A->bnnz;i++)
				{
					#ifdef _MPI
						fprintf(file, "%10d ",A->g_index[i]);
					#else
						fprintf(file, "%10d ",A->bindex[i]);
					#endif
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( A->bnnz%7!=0 ) fprintf(file, "\n");
				// value
				for(i=0;i<A->bnnz*A->bnr*A->bnc;i++)
				{
					fprintf(file, "%28.20e ",A->value[i]);
					if( (i+1)%3==0 ) fprintf(file, "\n");
				}
				if( (A->bnnz*A->bnr*A->bnc)%3!=0 ) fprintf(file, "\n");
				// RHS
				if( b!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",b->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				// solution
				if( x!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",x->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}

	}
	else
	{
		format = 1;
		format = *(char *)&format;
		// pointer
		tbuf   = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		tbuf[0]  = 10 + 3*sizeof(int) + (nprocs+1)*sizeof(size_t);
		size     = 6*sizeof(int);
		size    += (A->nc+1+A->bnnz)*sizeof(int) + A->bnnz*A->bnr*A->bnc*sizeof(LIS_SCALAR);
		if( b!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		if( x!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		#ifdef _MPI
		MPI_Gather(&size,sizeof(size_t)/sizeof(int),MPI_INT,&tbuf[1],sizeof(size_t)/sizeof(int),MPI_INT,0,A->comm);
		for(i=0;i<nprocs;i++) tbuf[i+1] += tbuf[i];
		#else
		tbuf[1] = size;
		#endif
		if( my_rank==0 )
		{
			file = fopen(path, "wb");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			if( format==LIS_BINARY_BIG )
			{
				sprintf(cbuf, "#LIS B %s", mattypename[A->matrix_type]);
			}
			else
			{
				sprintf(cbuf, "#LIS L %s", mattypename[A->matrix_type]);
			}
			fwrite(cbuf, sizeof(char), 10, file);
			fwrite(ibuf, sizeof(int), 3, file);
			fwrite(tbuf, sizeof(size_t), nprocs+1, file);
			fclose(file);
		}
		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "ab");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				ibuf[0] = A->n;
				ibuf[1] = A->nr;
				ibuf[2] = A->nc;
				ibuf[3] = A->bnr;
				ibuf[4] = A->bnc;
				ibuf[5] = A->bnnz;
				count = fwrite(ibuf, sizeof(int), 6, file);
				// bptr
				count = fwrite(A->bptr, sizeof(int), A->nc+1, file);
				// bindex
				#ifdef _MPI
					count = fwrite(A->g_index, sizeof(int), A->bnnz, file);
				#else
					count = fwrite(A->bindex, sizeof(int), A->bnnz, file);
				#endif
				// value
				count = fwrite(A->value, sizeof(LIS_SCALAR), A->bnnz*A->bnr*A->bnc, file);
				// RHS
				if( b!=NULL ) count = fwrite(b->value, sizeof(LIS_SCALAR), A->n, file);
				// solution
				if( x!=NULL ) count = fwrite(x->value, sizeof(LIS_SCALAR), A->n, file);
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}
		lis_free(tbuf);
	}
	return LIS_SUCCESS;
}

int lis_output_lis_vbr(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, char *path)
{
	int				n,nprocs,my_rank;
	int				i,pe;
	int				err;
	int				count;
	int				*ibuf;
	char			cbuf[128];
	char			buf[128];
	size_t			*tbuf,size;
	FILE			*file;


	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	n         = A->n;

	ibuf = (int *)buf;
	ibuf[0] = ibuf[1] = 0;
	if( b!=NULL ) ibuf[0] = 1;
	if( x!=NULL ) ibuf[1] = 1;
	ibuf[2] = A->nprocs;
	if( format==LIS_FMT_LIS_ASCII )
	{
		if( my_rank==0 )
		{
			file = fopen(path, "w");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}

			fprintf(file, "#LIS A %s\n", mattypename[A->matrix_type]);
			fprintf(file, "%d %d %d\n", ibuf[0], ibuf[1], ibuf[2]);
			fclose(file);
		}

		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "a");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				fprintf(file, "# %d %d %d %d %d %d\n", pe, A->n, A->nnz, A->nr, A->nc, A->bnnz);
				// row
				for(i=0;i<=A->nr;i++)
				{
					fprintf(file, "%10d ",A->row[i]);
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( (A->nr+1)%7!=0 ) fprintf(file, "\n");
				// col
				for(i=0;i<=A->nc;i++)
				{
					#ifdef _MPI
						fprintf(file, "%10d ",A->g_col[i]);
					#else
						fprintf(file, "%10d ",A->col[i]);
					#endif
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( (A->nc+1)%7!=0 ) fprintf(file, "\n");
				// ptr
				for(i=0;i<=A->bnnz;i++)
				{
					fprintf(file, "%10d ",A->ptr[i]);
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( (A->bnnz+1)%7!=0 ) fprintf(file, "\n");
				// bptr
				for(i=0;i<=A->nr;i++)
				{
					fprintf(file, "%10d ",A->bptr[i]);
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( (A->nr+1)%7!=0 ) fprintf(file, "\n");
				// bindex
				for(i=0;i<A->bnnz;i++)
				{
					#ifdef _MPI
						fprintf(file, "%10d ",A->g_index[i]);
					#else
						fprintf(file, "%10d ",A->bindex[i]);
					#endif
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( A->bnnz%7!=0 ) fprintf(file, "\n");
				// value
				for(i=0;i<A->nnz;i++)
				{
					fprintf(file, "%28.20e ",A->value[i]);
					if( (i+1)%3==0 ) fprintf(file, "\n");
				}
				if( A->nnz%3!=0 ) fprintf(file, "\n");
				// RHS
				if( b!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",b->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				// solution
				if( x!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",x->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}

	}
	else
	{
		format = 1;
		format = *(char *)&format;
		// pointer
		tbuf   = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		tbuf[0]  = 10 + 3*sizeof(int) + (nprocs+1)*sizeof(size_t);
		size     = 5*sizeof(int);
		size    += (A->nr+1+A->nr+1+A->nc+1+A->bnnz+A->bnnz+1)*sizeof(int) + A->nnz*sizeof(LIS_SCALAR);
		if( b!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		if( x!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		#ifdef _MPI
		MPI_Gather(&size,sizeof(size_t)/sizeof(int),MPI_INT,&tbuf[1],sizeof(size_t)/sizeof(int),MPI_INT,0,A->comm);
		for(i=0;i<nprocs;i++)
		{
			tbuf[i+1] += tbuf[i];
		}
		#else
		tbuf[1] = size;
		#endif
		if( my_rank==0 )
		{
			file = fopen(path, "wb");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			if( format==LIS_BINARY_BIG )
			{
				sprintf(cbuf, "#LIS B %s", mattypename[A->matrix_type]);
			}
			else
			{
				sprintf(cbuf, "#LIS L %s", mattypename[A->matrix_type]);
			}
			fwrite(cbuf, sizeof(char), 10, file);
			fwrite(ibuf, sizeof(int), 3, file);
			fwrite(tbuf, sizeof(size_t), nprocs+1, file);
			fclose(file);
		}
		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "ab");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				ibuf[0] = A->n;
				ibuf[1] = A->nnz;
				ibuf[2] = A->nr;
				ibuf[3] = A->nc;
				ibuf[4] = A->bnnz;
				count = fwrite(ibuf, sizeof(int), 5, file);
				// row
				count = fwrite(A->row, sizeof(int), A->nr+1, file);
				// col
				#ifdef _MPI
					count = fwrite(A->g_col, sizeof(int), A->nc+1, file);
				#else
					count = fwrite(A->col, sizeof(int), A->nc+1, file);
				#endif
				// ptr
				count = fwrite(A->ptr, sizeof(int), A->bnnz+1, file);
				// bptr
				count = fwrite(A->bptr, sizeof(int), A->nr+1, file);
				// bindex
				#ifdef _MPI
					count = fwrite(A->g_index, sizeof(int), A->bnnz, file);
				#else
					count = fwrite(A->bindex, sizeof(int), A->bnnz, file);
				#endif
				// value
				count = fwrite(A->value, sizeof(LIS_SCALAR), A->nnz, file);
				// RHS
				if( b!=NULL ) count = fwrite(b->value, sizeof(LIS_SCALAR), A->n, file);
				// solution
				if( x!=NULL ) count = fwrite(x->value, sizeof(LIS_SCALAR), A->n, file);
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}
		lis_free(tbuf);
	}
	return LIS_SUCCESS;
}

int lis_output_lis_coo(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, char *path)
{
	int				n,nprocs,my_rank;
	int				i,pe,is,nnz;
	int				err;
	int				count;
	int				*ibuf;
	char			cbuf[128];
	char			buf[128];
	size_t			*tbuf,size;
	FILE			*file;


	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	n         = A->n;
	nnz       = A->nnz;
	is        = A->is;

	ibuf = (int *)buf;
	ibuf[0] = ibuf[1] = 0;
	if( b!=NULL ) ibuf[0] = 1;
	if( x!=NULL ) ibuf[1] = 1;
	ibuf[2] = A->nprocs;
	if( format==LIS_FMT_LIS_ASCII )
	{
		if( my_rank==0 )
		{
			file = fopen(path, "w");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}

			fprintf(file, "#LIS A %s\n", mattypename[A->matrix_type]);
			fprintf(file, "%d %d %d\n", ibuf[0], ibuf[1], ibuf[2]);
			fclose(file);
		}

		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "a");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				fprintf(file, "# %d %d %d\n", pe, A->n, A->nnz);
				// row
				for(i=0;i<A->nnz;i++)
				{
					fprintf(file, "%10d ",A->row[i]+is);
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( A->nnz%7!=0 ) fprintf(file, "\n");
				// col
				for(i=0;i<A->nnz;i++)
				{
					#ifdef _MPI
					fprintf(file, "%10d ",A->g_index[i]);
					#else
					fprintf(file, "%10d ",A->col[i]);
					#endif
					if( (i+1)%7==0 ) fprintf(file, "\n");
				}
				if( A->nnz%7!=0 ) fprintf(file, "\n");
				// value
				for(i=0;i<A->nnz;i++)
				{
					fprintf(file, "%28.20e ",A->value[i]);
					if( (i+1)%3==0 ) fprintf(file, "\n");
				}
				if( A->nnz%3!=0 ) fprintf(file, "\n");
				// RHS
				if( b!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",b->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				// solution
				if( x!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",x->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}

	}
	else
	{
		format = 1;
		format = *(char *)&format;
		// pointer
		tbuf   = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		tbuf[0]  = 10 + 3*sizeof(int) + (nprocs+1)*sizeof(size_t);
		size     = 2*sizeof(int);
		size    += 2*A->nnz*sizeof(int) + A->nnz*sizeof(LIS_SCALAR);
		if( b!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		if( x!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		#ifdef _MPI
		MPI_Gather(&size,sizeof(size_t)/sizeof(int),MPI_INT,&tbuf[1],sizeof(size_t)/sizeof(int),MPI_INT,0,A->comm);
		for(i=0;i<nprocs;i++) tbuf[i+1] += tbuf[i];
		#else
		tbuf[1] = size;
		#endif
		if( my_rank==0 )
		{
			file = fopen(path, "wb");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			if( format==LIS_BINARY_BIG )
			{
				sprintf(cbuf, "#LIS B %s", mattypename[A->matrix_type]);
			}
			else
			{
				sprintf(cbuf, "#LIS L %s", mattypename[A->matrix_type]);
			}
			fwrite(cbuf, sizeof(char), 10, file);
			fwrite(ibuf, sizeof(int), 3, file);
			fwrite(tbuf, sizeof(size_t), nprocs+1, file);
			fclose(file);
		}
		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "ab");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				ibuf[0] = A->n;
				ibuf[1] = A->nnz;
				count = fwrite(ibuf, sizeof(int), 2, file);
				// row
				#ifdef _MPI
				for(i=0;i<nnz;i++) A->row[i] += is;
				#endif
				count = fwrite(A->row, sizeof(int), A->nnz, file);
				#ifdef _MPI
				for(i=0;i<nnz;i++) A->row[i] -= is;
				#endif
				// col
				#ifdef _MPI
				count = fwrite(A->g_index, sizeof(int), A->nnz, file);
				#else
				count = fwrite(A->col, sizeof(int), A->nnz, file);
				#endif
				// value
				count = fwrite(A->value, sizeof(LIS_SCALAR), A->nnz, file);
				// RHS
				if( b!=NULL ) count = fwrite(b->value, sizeof(LIS_SCALAR), A->n, file);
				// solution
				if( x!=NULL ) count = fwrite(x->value, sizeof(LIS_SCALAR), A->n, file);
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}
		lis_free(tbuf);
	}
	return LIS_SUCCESS;
}

int lis_output_lis_dns(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, char *path)
{
	int				n,gn,nprocs,my_rank;
	int				i,j,k,pe;
	int				err;
	int				count;
	int				*ibuf;
	int				*iw;
	char			cbuf[128];
	char			buf[128];
	size_t			*tbuf,size;
	FILE			*file;


	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	n         = A->n;
	gn        = A->gn;

#ifdef _MPI
	iw = (int *)lis_malloc( gn*sizeof(int) );
	for(i=0;i<gn;i++)
	{
		iw[A->g_index[i]] = i;
	}
#endif

	ibuf = (int *)buf;
	ibuf[0] = ibuf[1] = 0;
	if( b!=NULL ) ibuf[0] = 1;
	if( x!=NULL ) ibuf[1] = 1;
	ibuf[2] = A->nprocs;
	if( format==LIS_FMT_LIS_ASCII )
	{
		if( my_rank==0 )
		{
			file = fopen(path, "w");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}

			fprintf(file, "#LIS A %s\n", mattypename[A->matrix_type]);
			fprintf(file, "%d %d %d\n", ibuf[0], ibuf[1], ibuf[2]);
			fclose(file);
		}

		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "a");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				fprintf(file, "# %d %d\n", pe, A->n);
				// value
				k = 0;
				for(j=0;j<gn;j++)
				{
					for(i=0;i<n;i++)
					{
						#ifdef _MPI
							fprintf(file, "%28.20e ",A->value[iw[j]*n + i]);
						#else
							fprintf(file, "%28.20e ",A->value[j*n + i]);
						#endif
						if( (k+1)%3==0 ) fprintf(file, "\n");
						k++;
					}
				}
				printf("\n");
				if( k%3!=0 ) fprintf(file, "\n");
				// RHS
				if( b!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",b->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				// solution
				if( x!=NULL )
				{
					for(i=0;i<A->n;i++)
					{
						fprintf(file, "%28.20e ",x->value[i]);
						if( (i+1)%3==0 ) fprintf(file, "\n");
					}
					if( A->n%3!=0 ) fprintf(file, "\n");
				}
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}

	}
	else
	{
		format = 1;
		format = *(char *)&format;
		// pointer
		tbuf   = (size_t *)lis_malloc( (nprocs+1)*sizeof(size_t) );
		tbuf[0]  = 10 + 3*sizeof(int) + (nprocs+1)*sizeof(size_t);
		size     = 1*sizeof(int);
		size    += A->gn*A->n*sizeof(LIS_SCALAR);
		if( b!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		if( x!=NULL ) size += A->n*sizeof(LIS_SCALAR);
		#ifdef _MPI
		MPI_Gather(&size,sizeof(size_t)/sizeof(int),MPI_INT,&tbuf[1],sizeof(size_t)/sizeof(int),MPI_INT,0,A->comm);
		for(i=0;i<nprocs;i++)
		{
			tbuf[i+1] += tbuf[i];
		}
		#else
		tbuf[1] = size;
		#endif
		if( my_rank==0 )
		{
			file = fopen(path, "wb");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				return LIS_FAILS;
			}
			if( format==LIS_BINARY_BIG )
			{
				sprintf(cbuf, "#LIS B %s", mattypename[A->matrix_type]);
			}
			else
			{
				sprintf(cbuf, "#LIS L %s", mattypename[A->matrix_type]);
			}
			fwrite(cbuf, sizeof(char), 10, file);
			fwrite(ibuf, sizeof(int), 3, file);
			fwrite(tbuf, sizeof(size_t), nprocs+1, file);
			fclose(file);
		}
		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				file = fopen(path, "ab");
				if( file==NULL )
				{
					SETERR("cannot open file %s\n", path);
					return LIS_FAILS;
				}
				// params
				ibuf[0] = A->n;
				count = fwrite(ibuf, sizeof(int), 1, file);
				// value
				for(j=0;j<gn;j++)
				{
					#ifdef _MPI
						count = fwrite(&A->value[iw[j]*n], sizeof(LIS_SCALAR), n, file);
					#else
						count = fwrite(&A->value[j*n], sizeof(LIS_SCALAR), n, file);
					#endif
				}
				// RHS
				if( b!=NULL ) count = fwrite(b->value, sizeof(LIS_SCALAR), A->n, file);
				// solution
				if( x!=NULL ) count = fwrite(x->value, sizeof(LIS_SCALAR), A->n, file);
				fclose(file);
			}
			#ifdef _MPI
			MPI_Barrier(A->comm);
			#endif
		}
		lis_free(tbuf);
	}
#ifdef _MPI
	lis_free(iw);
#endif
	return LIS_SUCCESS;
}

int lis_output_vector(LIS_VECTOR v, int format, char *filename)
{
	int				err;
	FILE			*file;

	if( v==NULL )
	{
		SETERR("vector is NULL\n");
		return LIS_FAILS;
	}
	switch( format )
	{
	case LIS_FMT_PLAIN:
		err = lis_output_vector_plain(v,filename);
		break;
	case LIS_FMT_MM:
		err = lis_output_vector_mm(v,filename);
		break;
	case LIS_FMT_LIS:
		err = lis_output_vector_lis_ascii(v,filename);
		break;
	default:
		SETERR("ill format option\n");
		return LIS_FAILS;
		break;
	}
	return err;
}

int lis_output_vector_plain(LIS_VECTOR v, char *path)
{
	int				n,nprocs,my_rank;
	int				i,pe;
	int				err,ret;
	FILE			*file;

#ifdef _MPI
	nprocs    = v->nprocs;
	my_rank   = v->my_rank;
	n         = v->n;

	err = 0;
	for(pe=0;pe<nprocs;pe++)
	{
		if( my_rank==pe )
		{
			if( my_rank==0 )
			{
				file = fopen(path, "w");
			}
			else
			{
				file = fopen(path, "a");
			}
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				err = 1;
			}
			for(i=0;i<n;i++)
			{
				fprintf(file, "%28.20e\n", v->value[i]);
			}
			fclose(file);
		}
		MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,v->comm);
		if( ret )
		{
			return LIS_FAILS;
		}
	}
	return LIS_SUCCESS;
#else
	n         = v->n;

	file = fopen(path, "w");
	if( file==NULL )
	{
		SETERR("cannot open file %s\n", path);
		return LIS_FAILS;
	}
	for(i=0;i<n;i++)
	{
		fprintf(file, "%28.20e\n", v->value[i]);
	}
	fclose(file);
	return LIS_SUCCESS;
#endif
}

int lis_output_vector_mm(LIS_VECTOR v, char *path)
{
	int				n,nprocs,my_rank;
	int				i,pe,is;
	int				err,ret;
	FILE			*file;

#ifdef _MPI
	nprocs    = v->nprocs;
	my_rank   = v->my_rank;
	n         = v->n;
	is        = v->is;

	err = 0;
	if( my_rank==0 )
	{
		file = fopen(path, "w");
		if( file==NULL )
		{
			SETERR("cannot open file %s\n", path);
			err = 1;
		}

		fprintf(file, "%%%%MatrixMarket vector coordinate real general\n");
		fprintf(file, "%d\n", v->gn);
		fclose(file);
	}
	MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,v->comm);
	if( ret )
	{
		return LIS_FAILS;
	}

	err = 0;
	for(pe=0;pe<nprocs;pe++)
	{
		if( my_rank==pe )
		{
			file = fopen(path, "a");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				err = 1;
			}
			else
			{
				for(i=0;i<n;i++)
				{
					fprintf(file, "%d %28.20e\n", i+is+1, v->value[i]);
				}
				fclose(file);
			}
		}
		MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,v->comm);
		if( ret )
		{
			return LIS_FAILS;
		}
	}
	return LIS_SUCCESS;
#else
	n         = v->n;
	is        = v->is;

	file = fopen(path, "w");
	if( file==NULL )
	{
		SETERR("cannot open file %s\n", path);
		return LIS_FAILS;
	}

	fprintf(file, "%%%%MatrixMarket vector coordinate real general\n");
	fprintf(file, "%d\n", v->gn);

	for(i=0;i<n;i++)
	{
		fprintf(file, "%d %28.20e\n", i+is+1, v->value[i]);
	}
	fclose(file);

	return LIS_SUCCESS;
#endif
}


int lis_output_vector_lis_ascii(LIS_VECTOR v, char *path)
{
	int				n,nprocs,my_rank;
	int				i,pe;
	int				err,ret;
	FILE			*file;

#ifdef _MPI
	nprocs    = v->nprocs;
	my_rank   = v->my_rank;
	n         = v->n;

	err = 0;
	if( my_rank==0 )
	{
		file = fopen(path, "w");
		if( file==NULL )
		{
			SETERR("cannot open file %s\n", path);
			err = 1;
		}

		fprintf(file, "#LIS A vec\n");
		fprintf(file, "%d\n", nprocs);
		fclose(file);
	}
	MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,v->comm);
	if( ret )
	{
		return LIS_FAILS;
	}

	err = 0;
	for(pe=0;pe<nprocs;pe++)
	{
		if( my_rank==pe )
		{
			file = fopen(path, "a");
			if( file==NULL )
			{
				SETERR("cannot open file %s\n", path);
				err = 1;
			}
			else
			{
				fprintf(file, "# %d %d\n", pe, v->n);
				for(i=0;i<n;i++)
				{
					fprintf(file, "%28.20e ",v->value[i]);
					if( (i+1)%3==0 ) fprintf(file, "\n");
				}
				if( n%3!=0 ) fprintf(file, "\n");
				fclose(file);
			}
		}
		MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,v->comm);
		if( ret )
		{
			return LIS_FAILS;
		}
	}
	return LIS_SUCCESS;
#else
	n         = v->n;

	file = fopen(path, "w");
	if( file==NULL )
	{
		SETERR("cannot open file %s\n", path);
		return LIS_FAILS;
	}

	fprintf(file, "#LIS A vec\n");
	fprintf(file, "1\n");

	fprintf(file, "# 0 %d\n", v->n);
	for(i=0;i<n;i++)
	{
		fprintf(file, "%28.20e ",v->value[i]);
		if( (i+1)%3==0 ) fprintf(file, "\n");
	}
	if( n%3!=0 ) fprintf(file, "\n");
	fclose(file);

	return LIS_SUCCESS;
#endif
}
