#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include "lis.h"
#include "lis_precon.h"

#if defined(USE_SAAMG)
extern void finit_data_creation_(void *c_data_creation_ptr_bar);
extern void finit_v_cycle_(void *c_v_cycle_ptr_bar);
extern void finit_clear_matrix_(void *c_clear_matrix_ptr_bar);

char *f_data_creation_ptr;
char *f_v_cycle_ptr;
char *f_clear_matrix_ptr;
void c_data_creation_ptr_bar(char *bar)
{
	f_data_creation_ptr = bar;
}
void c_v_cycle_ptr_bar(char *bar)
{
	f_v_cycle_ptr = bar;
}
void c_clear_matrix_ptr_bar(char *bar)
{
	f_clear_matrix_ptr = bar;
}
#endif

int lis_precon_create(LIS_MATRIX A, LIS_VECTOR b, int precon_type, LIS_SCALAR params[], int options[], LIS_PRECON *precon)
{
	int err;
	int i,n,k;

	n = A->n;
	*precon = NULL;
	if( NULL==A )
	{
		SETERR("matrix A is undefined\n");
		return LIS_FAILS;
	}
	*precon = lis_malloc( sizeof(struct LIS_PRECON_STRUCT) );
	if( NULL==*precon )
	{
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}


	(*precon)->precon_type = precon_type;
	(*precon)->Amat        = NULL;
	(*precon)->Pmat        = NULL;
	(*precon)->L           = NULL;
	(*precon)->U           = NULL;
	(*precon)->Pb          = NULL;
	(*precon)->D           = NULL;
	(*precon)->temp        = NULL;
	(*precon)->work        = NULL;
	(*precon)->WS          = NULL;
	(*precon)->WR          = NULL;
	(*precon)->diag        = NULL;
	(*precon)->options     = NULL;
	(*precon)->params      = NULL;
	(*precon)->solver_comm = A->comm;

	if( precon_type==LIS_PRECON_TYPE_NONE )
	{
	    (*precon)->Amat    = A;
	    (*precon)->is_copy = LIS_FALSE;
	    return LIS_SUCCESS;
	}
	(*precon)->diag = lis_malloc( n*sizeof(int) );
	if( (*precon)->diag==NULL )
	{
		lis_precon_destroy(*precon);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	if( A->matrix_type!=LIS_MATRIX_CRS )
	{
		err = lis_matrix_convert(A, &(*precon)->Amat, LIS_MATRIX_CRS, NULL);
		if( err )
		{
			lis_precon_destroy(*precon);
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		(*precon)->is_copy = LIS_TRUE;
	}
	else
	{
		(*precon)->Amat    = A;
		(*precon)->is_copy = LIS_FALSE;
	}

	for(i=0;i<n;i++)
	{
		lis_sort_id(A->ptr[i],A->ptr[i+1]-1,A->index,A->value);
	}

	(*precon)->options = (int *)lis_malloc( LIS_OPTIONS_LEN*sizeof(int) );
	if( (*precon)->options==NULL )
	{
		lis_precon_destroy(*precon);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	(*precon)->params   = (LIS_SCALAR *)lis_malloc( LIS_PARAMS_LEN*sizeof(LIS_SCALAR) );
	if( (*precon)->params==NULL )
	{
		lis_precon_destroy(*precon);
		SETERR("out of memory\n");
		return LIS_OUT_OF_MEMORY;
	}
	memcpy((*precon)->options, options, LIS_OPTIONS_LEN*sizeof(int));
	memcpy((*precon)->params, params, LIS_PARAMS_LEN*sizeof(LIS_SCALAR));

	switch( precon_type )
	{
	case LIS_PRECON_TYPE_JACOBI:
		err = lis_vector_duplicate(b,&(*precon)->D);
		if( err )
		{
			lis_precon_destroy(*precon);
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		err = lis_precon_create_jacobi(*precon);
		break;
	case LIS_PRECON_TYPE_SSOR:
		err = lis_vector_duplicate(b,&(*precon)->D);
		if( err )
		{
			lis_precon_destroy(*precon);
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		err = lis_precon_create_ssor(*precon);
		break;
	case LIS_PRECON_TYPE_ILU:
		err = lis_vector_duplicate(b,&(*precon)->D);
		if( err )
		{
			lis_precon_destroy(*precon);
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		k = (*precon)->options[LIS_OPTIONS_FILL];
		if( k==0 )
		{
			err = lis_precon_create_ilu0(*precon);
		}
		else
		{
			err = lis_precon_create_iluk(*precon);
		}
		break;
/*
	case LIS_PRECON_TYPE_RIC2S:
			err = lis_precon_create_ric2s(*precon);
		break;
*/
	case LIS_PRECON_TYPE_SAI:
		err = lis_vector_duplicate(b,&(*precon)->D);
		if( err )
		{
			lis_precon_destroy(*precon);
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		err = lis_vector_duplicate(b,&(*precon)->temp);
		if( err )
		{
			lis_precon_destroy(*precon);
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		err = lis_precon_create_sainv(*precon);
		break;
	case LIS_PRECON_TYPE_IS:
		err = lis_vector_duplicate(b,&(*precon)->temp);
		if( err )
		{
			lis_precon_destroy(*precon);
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		k = (*precon)->options[LIS_OPTIONS_ISLEVEL];
		if( k==0 )
		{
			err = lis_precon_create_is(precon,b);
		}
		else
		{
			err = lis_vector_duplicate(b,&(*precon)->D);
			if( err )
			{
				lis_precon_destroy(*precon);
				SETERR("out of memory\n");
				return LIS_OUT_OF_MEMORY;
			}
			err = lis_precon_create_is2(*precon);
		}
		break;
	case LIS_PRECON_TYPE_HYBRID:
		err = lis_precon_create_iter(*precon);
		break;
#if defined(USE_SAAMG)
	case LIS_PRECON_TYPE_SAAMG:
		err = lis_vector_duplicateM(A,&(*precon)->temp);
		if( err )
		{
			lis_precon_destroy(*precon);
			SETERR("out of memory\n");
			return LIS_OUT_OF_MEMORY;
		}
		finit_data_creation_(c_data_creation_ptr_bar);
		finit_v_cycle_(c_v_cycle_ptr_bar);
		finit_clear_matrix_(c_clear_matrix_ptr_bar);
		lis_matrix_split(A,&(*precon)->D,&(*precon)->L,&(*precon)->U);
		for(i=A->n;i<A->np;i++) (*precon)->D->value[i] = 4.0;
		for(i=0;i<(*precon)->L->nnz;i++)
		{
			(*precon)->L->index[i]++;
		}
		for(i=0;i<(*precon)->U->nnz;i++)
		{
			(*precon)->U->index[i]++;
		}
#ifdef _MPI
		for(i=0;i<A->commtable->imnnz;i++)
		{
			A->commtable->node_import[i]++;
			A->commtable->node_export[i]++;
		}
		(*(void (*)())f_data_creation_ptr)(&A->n,&A->np,&A->commtable->imnnz,&(*precon)->L->nnz,&(*precon)->U->nnz,
			(*precon)->D->value,(*precon)->L->value,(*precon)->L->ptr,(*precon)->L->index,
			(*precon)->U->value, (*precon)->U->ptr, (*precon)->U->index,
			&A->commtable->neibpetot, A->commtable->neibpe, A->commtable->index_import,
			A->commtable->node_import, A->commtable->index_export, A->commtable->node_export,
			&A->comm, &(*precon)->level_num,&(*precon)->wsize);
		(*precon)->WS = (LIS_SCALAR *)lis_malloc((*precon)->wsize);
		(*precon)->WR = (LIS_SCALAR *)lis_malloc((*precon)->wsize);
#else
		(*(void (*)())f_data_creation_ptr)(&A->n,&(*precon)->L->nnz,&(*precon)->U->nnz,
			(*precon)->D->value,(*precon)->L->value,(*precon)->L->ptr,(*precon)->L->index,
			(*precon)->U->value, (*precon)->U->ptr, (*precon)->U->index, &(*precon)->level_num);
#endif
		err = LIS_SUCCESS;
	   break;
#endif
	}

	if( err )
	{
		lis_precon_destroy(*precon);
		return err;
	}
	return LIS_SUCCESS;
}

void lis_precon_destroy(LIS_PRECON precon)
{
	int i;

	if( precon )
	{
		lis_vector_destroy(precon->Pb);
		lis_vector_destroy(precon->D);
		lis_vector_destroy(precon->temp);
		if( precon->work    )
		{
			for(i=0;i<precon->worklen;i++)
			{
				lis_vector_destroy(precon->work[i]);
			}
			lis_free(precon->work);
		}
		if( precon->diag    ) lis_free(precon->diag);
		if( precon->WS      ) lis_free(precon->WS);
		if( precon->WR      ) lis_free(precon->WR);
		if( precon->options ) lis_free(precon->options);
		if( precon->params  ) lis_free(precon->params);
		if( precon->is_copy ) lis_matrix_destroy(precon->Amat);
		lis_matrix_destroy(precon->Pmat);
		lis_matrix_destroy(precon->L);
		lis_matrix_destroy(precon->U);
#if defined(USE_SAAMG)
		if( precon->precon_type==LIS_PRECON_TYPE_SAAMG )
		{
		 (*(void (*)())f_clear_matrix_ptr)(&precon->level_num);
		}
#endif
		lis_free(precon);
	}
}

void lis_psolve(LIS_PRECON precon, LIS_VECTOR b, LIS_VECTOR x)
{
	int k,n;

	if( precon==NULL )
	{
		lis_vector_copy(b,x);
		return;
	}

	switch( precon->precon_type )
	{
	case LIS_PRECON_TYPE_JACOBI:
		lis_psolve_jacobi(precon,b,x);
		break;
	case LIS_PRECON_TYPE_SSOR:
		lis_psolve_ssor(precon,b,x);
		break;
	case LIS_PRECON_TYPE_ILU:
		k = precon->options[LIS_OPTIONS_FILL];
		if( k==0 )
		{
			lis_psolve_ilu0(precon,b,x);
		}
		else
		{
			lis_psolve_iluk(precon,b,x);
		}
		break;
	case LIS_PRECON_TYPE_IS:
		k = precon->options[LIS_OPTIONS_ISLEVEL];
		if( k==0 )
		{
			lis_psolve_is(precon,b,x);
		}
		else if( k==1 )
		{
			lis_psolve_is_1(precon,b,x);
		}
		else
		{
			lis_psolve_is_2(precon,b,x);
		}
		break;
	case LIS_PRECON_TYPE_SAI:
		lis_psolve_sainv(precon,b,x);
		break;
#if defined(USE_SAAMG)
	case LIS_PRECON_TYPE_SAAMG:
		lis_vector_set_all(0,x);
#ifdef _MPI
		n = b->np;
		 (*(void (*)())f_v_cycle_ptr)(&n,b->value,x->value,precon->temp->value,&precon->level_num,
			 &precon->solver_comm,precon->WS,precon->WR,&precon->wsize);
#else
		n = b->n;
		 (*(void (*)())f_v_cycle_ptr)(&n,b->value,x->value,&precon->level_num,precon->temp->value);
#endif
		break;
#endif
	case LIS_PRECON_TYPE_HYBRID:
		lis_psolve_iter(precon,b,x);
		break;
/*
	case LIS_PRECON_TYPE_RIC2S:
		lis_psolve_ric2s(precon,b,x);
		break;
*/
	case LIS_PRECON_TYPE_NONE:
	default:
		lis_vector_copy(b, x);
		break;
	}
	return;
}

void lis_psolvet(LIS_PRECON precon, LIS_VECTOR b, LIS_VECTOR x)
{
	int k,n;

	if( precon==NULL )
	{
		lis_vector_copy(b,x);
	}

	switch( precon->precon_type )
	{
	case LIS_PRECON_TYPE_JACOBI:
		lis_psolvet_jacobi(precon,b,x);
		break;
	case LIS_PRECON_TYPE_SSOR:
		lis_psolvet_ssor(precon,b,x);
		break;
	case LIS_PRECON_TYPE_ILU:
		k = precon->options[LIS_OPTIONS_FILL];
		if( k==0 )
		{
			lis_psolvet_ilu0(precon,b,x);
		}
		else
		{
			lis_psolvet_iluk(precon,b,x);
		}
		break;
	case LIS_PRECON_TYPE_IS:
		k = precon->options[LIS_OPTIONS_ISLEVEL];
		if( k==0 )
		{
			lis_psolvet_is(precon,b,x);
		}
		else if( k==1 )
		{
			lis_psolvet_is_1(precon,b,x);
		}
		else
		{
			lis_psolvet_is_2(precon,b,x);
		}
		break;
	case LIS_PRECON_TYPE_HYBRID:
		lis_psolvet_iter(precon,b,x);
		break;
	case LIS_PRECON_TYPE_SAI:
		lis_psolvet_sainv(precon,b,x);
		break;
#if defined(USE_SAAMG)
	case LIS_PRECON_TYPE_SAAMG:
		lis_vector_set_all(0,x);
#ifdef _MPI
		n = b->np;
		 (*(void (*)())f_v_cycle_ptr)(&n,b->value,x->value,precon->temp->value,&precon->level_num,
			 &precon->solver_comm,precon->WS,precon->WR,&precon->wsize);
#else
		n = b->n;
		 (*(void (*)())f_v_cycle_ptr)(&n,b->value,x->value,&precon->level_num,precon->temp->value);
#endif
		break;
#endif
/*
	case LIS_PRECON_TYPE_RIC2S:
			lis_psolvet_ric2s(precon,b,x);
		break;
*/
	case LIS_PRECON_TYPE_NONE:
	default:
		lis_vector_copy(b, x);
		break;
	}
	return;
}


