#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <malloc.h>
#include "lis.h"
#include "lis_solver.h"

int lis_cg(LIS_VECTOR b, LIS_VECTOR x, LIS_VECTOR work[],
		   LIS_SCALAR params[], int options[],
		   LIS_MATRIX A, LIS_PRECON M, LIS_SCALAR status[])
{
	LIS_VECTOR r,z,p,q;
	LIS_SCALAR alpha, beta, rho, rho_old, rhotol;
	LIS_SCALAR bnrm2, nrm2, tol, dot_pq, nrm2bak;
	int iter,maxiter,n,np,output,i;
	double ptimes, times;

	n  = A->n;
	maxiter = options[LIS_OPTIONS_MAXITER];
	output  = options[LIS_OPTIONS_OUTPUT];
	tol     = params[LIS_PARAMS_RESID];
	ptimes  = 0.0;

	
	z = work[0];
	q = work[1];
	r = work[2];
	p = work[3];
	rho_old = (LIS_SCALAR)1.0;
	beta = (LIS_SCALAR)0.0;

	// Initial Residual
	if( lis_get_initial_residual(A,NULL,x,b,r,NULL,tol,&bnrm2,status) )
	{
		return LIS_SUCCESS;
	}
	lis_vector_set_all(0.0,p);
	
	for( iter=1; iter<=maxiter; iter++ )
	{
		/* z = M^-1 * r */
		times = lis_wtime();
		lis_psolve(M,r,z);
		ptimes += lis_wtime() - times;

		/* rho = <r,z> */
		lis_vector_dot(r,z,&rho);

		/* beta = rho / rho_old */
		beta = rho / rho_old;

		/* p = z + beta*p       */
		lis_vector_xpay(z,beta,p);
		
		/* q = Ap */
		lis_matvec(A,p,q);
		
		/* dot_pq = <p,q> */
		lis_vector_dot(p,q,&dot_pq);

		/* breakdown check */
		if( dot_pq==0.0 )
		{
			status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_BREAKDOWN;
			status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
			status[LIS_STATUS_RESID] = nrm2;
			return LIS_BREAKDOWN;
		}
		
		/* alpha = rho / dot_pq */
		alpha = rho / dot_pq;
		
		/* x = x + alpha*p */
		lis_vector_axpy(alpha,p,x);
		
		/* r = r - alpha*q */
		lis_vector_axpy(-alpha,q,r);

		/* convergence check */
		lis_vector_nrm2(r,&nrm2);
		nrm2 = nrm2 * bnrm2;

		if( output )
		{
			if( output & LIS_PRINT_MEM ) lis_set_residual_history(nrm2);
			if( output & LIS_PRINT_OUT ) printf("iter: %5d  residual = %e\n", iter, nrm2);
		}
		
		if( tol >= nrm2 )
		{
			status[LIS_STATUS_RET]    = (LIS_SCALAR)LIS_SUCCESS;
			status[LIS_STATUS_ITER]   = (LIS_SCALAR)iter;
			status[LIS_STATUS_RESID]  = nrm2;
			status[LIS_STATUS_PTIMES] = ptimes;
			return LIS_SUCCESS;
		}
		rho_old = rho;
	}

	status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_MAXITER;
	status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
	status[LIS_STATUS_RESID] = nrm2;
	return LIS_MAXITER;
}

