#include <math.h>
#include <stdlib.h>
#include <malloc.h>
#include "lis.h"
#include "lis_solver.h"


int lis_sor(LIS_VECTOR B, LIS_VECTOR X, LIS_VECTOR work[],
		   LIS_SCALAR params[], int options[],
		   LIS_MATRIX A, LIS_PRECON M, LIS_SCALAR status[])
{
	LIS_SCALAR ux,lx;
	LIS_SCALAR bb,xt;
	LIS_SCALAR nrm2;
	LIS_SCALAR tol;
	LIS_SCALAR w;
	LIS_SCALAR *x,*b;
	LIS_SCALAR *r,*d,*rr;
	LIS_VECTOR RR;
	int iter,maxiter,n,output;
	int i,j;
	int *diag;

	n       = A->n;
	maxiter = options[LIS_OPTIONS_MAXITER];
	output  = options[LIS_OPTIONS_OUTPUT];
	tol     = params[LIS_PARAMS_RESID];
	w       = params[LIS_PARAMS_W];

	diag    = (int *)lis_malloc( n*sizeof(int) );

	RR      = work[2];
	x       = X->value;
	b       = B->value;
	r       = work[0]->value;
	d       = work[1]->value;
	rr      = work[2]->value;

	lis_vector_nrm2(B,&bb);
	bb      = 1.0 / bb;
#pragma omp parallel for private(i,j)
	for(i=0;i<n;i++)
	{
		r[i] = b[i];
	}
#pragma omp parallel for private(i,j)
	for(i=0; i<n; i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			if( i==A->index[j] )
			{
				diag[i] = j;
				d[i]    = 1.0 / A->value[j];
				break;
			}
		}
	}

	for( iter=1; iter<=maxiter; iter++ )
	{
//#pragma omp parallel for private(i,j,lx,ux)
#ifdef _MPI
		if( A->np>X->np )
		{
			X->value = (LIS_SCALAR *)lis_realloc(X->value,A->np*sizeof(LIS_SCALAR));
			if( X->value==NULL )
			{
				SETERR("out of memory\n");
				return LIS_OUT_OF_MEMORY;
			}
			x = X->value;
		}
		lis_send_recv(A->commtable,x);
#endif
#pragma omp parallel for private(i,j,lx,ux,xt)
		for(i=0;i<n;i++)
		{
			lx = b[i];
			ux = 0.0;
			for(j=A->ptr[i];j<diag[i];j++)
			{
				lx -= A->value[j] * x[A->index[j]];
			}
			for(j=diag[i]+1;j<A->ptr[i+1];j++)
			{
				ux -= A->value[j] * x[A->index[j]];
			}
			rr[i]  = r[i] + ux - A->value[diag[i]]*x[i];
			xt       = (lx + ux) * d[i];
			x[i]     = x[i] + w*(xt - x[i]);
			r[i]   = lx;
		}

		/* convergence check */
		lis_vector_nrm2(RR,&nrm2);
		nrm2 = nrm2 * bb;

		if( output )
		{
			if( output & LIS_PRINT_MEM ) lis_set_residual_history(nrm2);
			if( output & LIS_PRINT_OUT ) printf("iter: %5d  residual = %e\n", iter, nrm2);
		}

		if( tol >= nrm2 )
		{
			status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_SUCCESS;
			status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
			status[LIS_STATUS_RESID] = nrm2;
			status[LIS_STATUS_PTIMES] = 0.0;
			lis_free(diag);
			return LIS_SUCCESS;
		}
	}

	status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_MAXITER;
	status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
	status[LIS_STATUS_RESID] = nrm2;
	lis_free(diag);
	return LIS_MAXITER;
}

