#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef _MPI
	#include <mpi.h>
#endif
#include "lis.h"


int main(int argc, char* argv[])
{
	LIS_MATRIX		A;
	LIS_VECTOR		b,x,u;

	int				nprocs,my_rank;
	int				i,k,n,gn;
	int				is,ie;
	int				err,iter;
	int				options[LIS_OPTIONS_LEN];

	char			solvername[128];

	LIS_SCALAR		params[LIS_PARAMS_LEN];
	LIS_SCALAR		status[LIS_STATUS_LEN];
	LIS_SCALAR		times,resid;
	FILE 			*infile,*resfile;

	lis_initialize(argc, argv, NULL, options, params);

	if( argc < 4 )
	{
		printf("test1 matrix_filename result_filename residual_filename [options]\n");
		lis_finalize();
		exit(0);
	}

#ifdef _MPI
    MPI_Comm_size(MPI_COMM_WORLD,&nprocs);
    MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
#else
	nprocs  = 1;
	my_rank = 0;
#endif

	#ifdef _OPENMP
	printf("Max Procs   = %d\n",omp_get_num_procs());
	printf("Max Threads = %d\n",omp_get_max_threads());
	#endif

	/* read matrix and vectors from file */
	err = lis_input(&A,&b,&x,argv[1],LIS_COMM_WORLD,LIS_MATRIX_CRS);
	CHKERR(err);
	n  = A->n;
	gn = A->gn;

	err = lis_vector_create(0,gn,LIS_COMM_WORLD,&u);
	CHKERR(err);
	if(b==NULL)
	{
		lis_vector_duplicate(u,&b);
		CHKERR(err);
		err = lis_vector_set_all(1.0,u);
		lis_matvec(A,u,b);
	}
	if(x==NULL)
	{
		err = lis_vector_duplicate(u,&x);
		CHKERR(err);
		err = lis_vector_set_all(0.0,x);
	}

	options[LIS_OPTIONS_OUTPUT] = LIS_PRINT_MEM;
	err = lis_solve(A,b,x,params,options,status);
	CHKERR(err);

	iter  = (int)status[LIS_STATUS_ITER];
	resid = status[LIS_STATUS_RESID];
	times = status[LIS_STATUS_TIMES];
	k     = options[LIS_OPTIONS_SOLVER];
	lis_get_solvername(k,solvername);
	if( my_rank==0 )
	{
		printf("%s: iter = %d time = %e (p=%e i=%e)\n",solvername,iter, times, status[LIS_STATUS_PTIMES], status[LIS_STATUS_ITIMES]);
		printf("%s: Last Residual = %e \n",solvername,resid);
	}

	/* write solution */
	lis_output_vector(x,LIS_FMT_MM,argv[2]);

	/* write residual */
	lis_output_residual_history(argv[3]);

	lis_matrix_destroy(A);
	lis_vector_destroy(b);
	lis_vector_destroy(x);
	lis_vector_destroy(u);

	lis_finalize();
	return 0;
}

